/* ARCADEmini simulator (unix, console)
 * version 0.5 date 2003-09-03
 * part of the ARCADEmini project - http[s]://arcademini.schuermans.info/
 * Copyright (C) 2003 sphaera <sphaera@geekness.de>
 *                and 1stein <1stein@schuermans.info>
 * Copyleft: GNU public license - http://www.gnu.org/copyleft/gpl.html
 */

//ARCADEmini device

#include <stdio.h>
#include <unistd.h>
#include <time.h>
#include <sys/time.h>

#include "am_dev.h"

//-------------------------------------------------------------------------------------------
//load a new sector
//-------------------------------------------------------------------------------------------
void am_dev_load_sector( st_am_state * p_state )
{
  long long pos;

  //default: clear sector buffer
  memset( p_state->sector, 0, sizeof( p_state->sector ) );

  //no file open
  if( p_state->p_file == NULL )
    return;

  //calculate position in file
  pos = (long long)p_state->sector_no << 9;

  //seek to position
  if( fseeko64( p_state->p_file, pos, SEEK_SET ) != 0 )
    return;

  //read sector
  fread( p_state->sector, 1, sizeof( p_state->sector ), p_state->p_file );

  //end of file
  if( feof( p_state->p_file ) )
  {
    //restart with second sector
    p_state->sector_no = 1;

    //seek to second sector
    if( fseeko64( p_state->p_file, 0x200, SEEK_SET ) != 0 )
      return;

    //read sector
    fread( p_state->sector, 1, sizeof( p_state->sector ), p_state->p_file );
  } 
}

//-------------------------------------------------------------------------------------------
// JMP_NEXT_SEC
// load the next sector
//-------------------------------------------------------------------------------------------
void am_dev_JMP_NEXT_SEC( st_am_state * p_state )
{
  //increase sector number
  p_state->sector_no++;
  //continue at start of sector
  p_state->offset = 0;
  //load new sector
  am_dev_load_sector( p_state );
}

//-------------------------------------------------------------------------------------------
// JMP_SEC <sectors (4)>
// jump <sectors> sectors further
// - <sectors (4)> is stored in big endian format
//-------------------------------------------------------------------------------------------
void am_dev_JMP_SEC( st_am_state * p_state )
{
  unsigned long sectors;
  //get number of sectors
  sectors = (unsigned long)(p_state->sector[p_state->offset++]) << 24;
  sectors |= (unsigned long)(p_state->sector[p_state->offset++]) << 16;
  sectors |= (unsigned long)(p_state->sector[p_state->offset++]) << 8;
  sectors |= (unsigned long)(p_state->sector[p_state->offset++]);
  //get absolute number of new secotr
  p_state->sector_no += sectors;
  //continue at start of sector
  p_state->offset = 0;
  //load new sector
  am_dev_load_sector( p_state );
}

//-------------------------------------------------------------------------------------------
// JMP_OFS (bit0 is bit8 of offset) <offset>
// jump to offset
//-------------------------------------------------------------------------------------------
void am_dev_JMP_OFS( st_am_state * p_state )
{
  unsigned char offset;
  //get offset
  offset = (unsigned short)(p_state->sector[p_state->offset-1] & 0x01) << 8;
  offset |= (unsigned short)(p_state->sector[p_state->offset++]);
  //jump to offset
  p_state->offset = offset;
}

//-------------------------------------------------------------------------------------------
// JMP_OFS_SEC (bit0 is bit8 of offset) <offset> <sectors (4)>
// jump to offset and <sectors> sectors further
// - <sectors (4)> is stored in big endian format
//-------------------------------------------------------------------------------------------
void am_dev_JMP_OFS_SEC( st_am_state * p_state )
{
  unsigned char offset;
  unsigned long sectors;
  //get offset
  offset = (unsigned short)(p_state->sector[p_state->offset-1] & 0x01) << 8;
  offset |= (unsigned short)(p_state->sector[p_state->offset++]);
  //get number of sectors
  sectors = (unsigned long)(p_state->sector[p_state->offset++]) << 24;
  sectors |= (unsigned long)(p_state->sector[p_state->offset++]) << 16;
  sectors |= (unsigned long)(p_state->sector[p_state->offset++]) << 8;
  sectors |= (unsigned long)(p_state->sector[p_state->offset++]);
  //get absolute number of new secotr
  p_state->sector_no += sectors;
  //jump to offset
  p_state->offset = offset;
  //load new sector
  am_dev_load_sector( p_state );
}

//-------------------------------------------------------------------------------------------
// JMP_IND <offset_h> <offset_l> <sectors_hh> <sectors_hl> <sectors_lh> <sectors_ll>
// jump indirect to offset and <sectors> sectors further
//-------------------------------------------------------------------------------------------
void am_dev_JMP_IND( st_am_state * p_state )
{
  unsigned char offset;
  unsigned long sectors;
  //get offset
  offset = (unsigned short)(p_state->sector[p_state->offset-1] & 0x01) << 8;
  offset |= (unsigned short)(p_state->sector[p_state->offset++]);
  //get number of sectors
  sectors = (unsigned long)(p_state->sector[p_state->offset++]) << 24;
  sectors |= (unsigned long)(p_state->sector[p_state->offset++]) << 16;
  sectors |= (unsigned long)(p_state->sector[p_state->offset++]) << 8;
  sectors |= (unsigned long)(p_state->sector[p_state->offset++]);
  //get absolute number of new secotr
  p_state->sector_no += sectors;
  //jump to offset
  p_state->offset = offset;
  //load new sector
  am_dev_load_sector( p_state );
}

//-------------------------------------------------------------------------------------------
// JMP_EQ (bit0 is bit8 of offset) <offset> <param1> <param2>
// jump if equal
// jump to offset if <param1> == <param2>
//-------------------------------------------------------------------------------------------
void am_dev_JMP_EQ( st_am_state * p_state )
{
  unsigned short offset;
  unsigned char param1, param2;
  //get offset
  offset = (unsigned short)(p_state->sector[p_state->offset-1] & 0x01) << 8;
  offset |= (unsigned short)(p_state->sector[p_state->offset++]);
  //get param1, param2
  param1 = p_state->vars[p_state->sector[p_state->offset++]];
  param2 = p_state->vars[p_state->sector[p_state->offset++]];
  //jump if equal
  if( param1 == param2 )
    p_state->offset = offset;
}

//-------------------------------------------------------------------------------------------
// JMP_EQ_C (bit0 is bit8 of offset) <offset> <param> <constant>
// jump if equal to constant
// jump to offset if <param> == <constant>
//-------------------------------------------------------------------------------------------
void am_dev_JMP_EQ_C( st_am_state * p_state )
{
  unsigned short offset;
  unsigned char param, constant;
  //get offset
  offset = (unsigned short)(p_state->sector[p_state->offset-1] & 0x01) << 8;
  offset |= (unsigned short)(p_state->sector[p_state->offset++]);
  //get param, constant
  param = p_state->vars[p_state->sector[p_state->offset++]];
  constant = p_state->sector[p_state->offset++];
  //jump if equal
  if( param == constant )
    p_state->offset = offset;
}

//-------------------------------------------------------------------------------------------
// JMP_NEQ (bit0 is bit8 of offset) <offset> <param1> <param2>
// jump if not equal
// jump to offset if <param1> != <param2>
//-------------------------------------------------------------------------------------------
void am_dev_JMP_NEQ( st_am_state * p_state )
{
  unsigned short offset;
  unsigned char param1, param2;
  //get offset
  offset = (unsigned short)(p_state->sector[p_state->offset-1] & 0x01) << 8;
  offset |= (unsigned short)(p_state->sector[p_state->offset++]);
  //get param1, param2
  param1 = p_state->vars[p_state->sector[p_state->offset++]];
  param2 = p_state->vars[p_state->sector[p_state->offset++]];
  //jump if not equal
  if( param1 != param2 )
    p_state->offset = offset;
}

//-------------------------------------------------------------------------------------------
// JMP_NEQ_C (bit0 is bit8 of offset) <offset> <param> <constant>
// jump if not equal to constant
// jump to offset if <param> != <constant>
//-------------------------------------------------------------------------------------------
void am_dev_JMP_NEQ_C( st_am_state * p_state )
{
  unsigned short offset;
  unsigned char param, constant;
  //get offset
  offset = (unsigned short)(p_state->sector[p_state->offset-1] & 0x01) << 8;
  offset |= (unsigned short)(p_state->sector[p_state->offset++]);
  //get param, constant
  param = p_state->vars[p_state->sector[p_state->offset++]];
  constant = p_state->sector[p_state->offset++];
  //jump if not equal
  if( param != constant )
    p_state->offset = offset;
}

//-------------------------------------------------------------------------------------------
// JMP_GT (bit0 is bit8 of offset) <offset> <param1> <param2>
// jump if greater than
// jump to offset if <param1> > <param2>
//-------------------------------------------------------------------------------------------
void am_dev_JMP_GT( st_am_state * p_state )
{
  unsigned short offset;
  unsigned char param1, param2;
  //get offset
  offset = (unsigned short)(p_state->sector[p_state->offset-1] & 0x01) << 8;
  offset |= (unsigned short)(p_state->sector[p_state->offset++]);
  //get param1, param2
  param1 = p_state->vars[p_state->sector[p_state->offset++]];
  param2 = p_state->vars[p_state->sector[p_state->offset++]];
  //jump if greater
  if( param1 > param2 )
    p_state->offset = offset;
}

//-------------------------------------------------------------------------------------------
// JMP_GT_C (bit0 is bit8 of offset) <offset> <param> <constant>
// jump if greater than
// jump to offset if <param> > <constant>
//-------------------------------------------------------------------------------------------
void am_dev_JMP_GT_C( st_am_state * p_state )
{
  unsigned short offset;
  unsigned char param, constant;
  //get offset
  offset = (unsigned short)(p_state->sector[p_state->offset-1] & 0x01) << 8;
  offset |= (unsigned short)(p_state->sector[p_state->offset++]);
  //get param, constant
  param = p_state->vars[p_state->sector[p_state->offset++]];
  constant = p_state->sector[p_state->offset++];
  //jump if greater
  if( param > constant )
    p_state->offset = offset;
}

//-------------------------------------------------------------------------------------------
// JMP_NGT (bit0 is bit8 of offset) <offset> <param1> <param2>
// jump if not greater than
// jump to offset if <param1> <= <param2>
//-------------------------------------------------------------------------------------------
void am_dev_JMP_NGT( st_am_state * p_state )
{
  unsigned short offset;
  unsigned char param1, param2;
  //get offset
  offset = (unsigned short)(p_state->sector[p_state->offset-1] & 0x01) << 8;
  offset |= (unsigned short)(p_state->sector[p_state->offset++]);
  //get param1, param2
  param1 = p_state->vars[p_state->sector[p_state->offset++]];
  param2 = p_state->vars[p_state->sector[p_state->offset++]];
  //jump if not greater
  if( param1 <= param2 )
    p_state->offset = offset;
}

//-------------------------------------------------------------------------------------------
// JMP_NGT_C (bit0 is bit8 of offset) <offset> <param> <constant>
// jump if not greater than
// jump to offset if <param> <= <constant>
//-------------------------------------------------------------------------------------------
void am_dev_JMP_NGT_C( st_am_state * p_state )
{
  unsigned short offset;
  unsigned char param, constant;
  //get offset
  offset = (unsigned short)(p_state->sector[p_state->offset-1] & 0x01) << 8;
  offset |= (unsigned short)(p_state->sector[p_state->offset++]);
  //get param, constant
  param = p_state->vars[p_state->sector[p_state->offset++]];
  constant = p_state->sector[p_state->offset++];
  //jump if not greater
  if( param <= constant )
    p_state->offset = offset;
}

//-------------------------------------------------------------------------------------------
// JMP_LT (bit0 is bit8 of offset) <offset> <param1> <param2>
// jump if less than
// jump to offset if <param1> < <param2>
//-------------------------------------------------------------------------------------------
void am_dev_JMP_LT( st_am_state * p_state )
{
  unsigned short offset;
  unsigned char param1, param2;
  //get offset
  offset = (unsigned short)(p_state->sector[p_state->offset-1] & 0x01) << 8;
  offset |= (unsigned short)(p_state->sector[p_state->offset++]);
  //get param1, param2
  param1 = p_state->vars[p_state->sector[p_state->offset++]];
  param2 = p_state->vars[p_state->sector[p_state->offset++]];
  //jump if less
  if( param1 < param2 )
    p_state->offset = offset;
}

//-------------------------------------------------------------------------------------------
// JMP_LT_C (bit0 is bit8 of offset) <offset> <param> <constant>
// jump if less than
// jump to offset if <param> < <constant>
//-------------------------------------------------------------------------------------------
void am_dev_JMP_LT_C( st_am_state * p_state )
{
  unsigned short offset;
  unsigned char param, constant;
  //get offset
  offset = (unsigned short)(p_state->sector[p_state->offset-1] & 0x01) << 8;
  offset |= (unsigned short)(p_state->sector[p_state->offset++]);
  //get param, constant
  param = p_state->vars[p_state->sector[p_state->offset++]];
  constant = p_state->sector[p_state->offset++];
  //jump if less
  if( param < constant )
    p_state->offset = offset;
}

//-------------------------------------------------------------------------------------------
// JMP_NLT (bit0 is bit8 of offset) <offset> <param1> <param2>
// jump if not less than
// jump to offset if <param1> >= <param2>
//-------------------------------------------------------------------------------------------
void am_dev_JMP_NLT( st_am_state * p_state )
{
  unsigned short offset;
  unsigned char param1, param2;
  //get offset
  offset = (unsigned short)(p_state->sector[p_state->offset-1] & 0x01) << 8;
  offset |= (unsigned short)(p_state->sector[p_state->offset++]);
  //get param1, param2
  param1 = p_state->vars[p_state->sector[p_state->offset++]];
  param2 = p_state->vars[p_state->sector[p_state->offset++]];
  //jump if not less
  if( param1 >= param2 )
    p_state->offset = offset;
}

//-------------------------------------------------------------------------------------------
// JMP_NLT_C (bit0 is bit8 of offset) <offset> <param> <constant>
// jump if not less than
// jump to offset if <param> >= <constant>
//-------------------------------------------------------------------------------------------
void am_dev_JMP_NLT_C( st_am_state * p_state )
{
  unsigned short offset;
  unsigned char param, constant;
  //get offset
  offset = (unsigned short)(p_state->sector[p_state->offset-1] & 0x01) << 8;
  offset |= (unsigned short)(p_state->sector[p_state->offset++]);
  //get param, constant
  param = p_state->vars[p_state->sector[p_state->offset++]];
  constant = p_state->sector[p_state->offset++];
  //jump if not less
  if( param >= constant )
    p_state->offset = offset;
}

//-------------------------------------------------------------------------------------------
// JMP_GTS (bit0 is bit8 of offset) <offset> <param1> <param2>
// jump if greater (signed) than
// jump to offset if <param1> > <param2>
//-------------------------------------------------------------------------------------------
void am_dev_JMP_GTS( st_am_state * p_state )
{
  unsigned short offset;
  unsigned char param1, param2;
  //get offset
  offset = (unsigned short)(p_state->sector[p_state->offset-1] & 0x01) << 8;
  offset |= (unsigned short)(p_state->sector[p_state->offset++]);
  //get param1, param2
  param1 = p_state->vars[p_state->sector[p_state->offset++]];
  param2 = p_state->vars[p_state->sector[p_state->offset++]];
  //jump if greater (signed)
  if( (signed char)param1 > (signed char)param2 )
    p_state->offset = offset;
}

//-------------------------------------------------------------------------------------------
// JMP_GTS_C (bit0 is bit8 of offset) <offset> <param> <constant>
// jump if greater (signed) than
// jump to offset if <param> > <constant>
//-------------------------------------------------------------------------------------------
void am_dev_JMP_GTS_C( st_am_state * p_state )
{
  unsigned short offset;
  unsigned char param, constant;
  //get offset
  offset = (unsigned short)(p_state->sector[p_state->offset-1] & 0x01) << 8;
  offset |= (unsigned short)(p_state->sector[p_state->offset++]);
  //get param, constant
  param = p_state->vars[p_state->sector[p_state->offset++]];
  constant = p_state->sector[p_state->offset++];
  //jump if greater (signed)
  if( (signed char)param > (signed char)constant )
    p_state->offset = offset;
}

//-------------------------------------------------------------------------------------------
// JMP_NGTS (bit0 is bit8 of offset) <offset> <param1> <param2>
// jump if not greater (signed) than
// jump to offset if <param1> <= <param2>
//-------------------------------------------------------------------------------------------
void am_dev_JMP_NGTS( st_am_state * p_state )
{
  unsigned short offset;
  unsigned char param1, param2;
  //get offset
  offset = (unsigned short)(p_state->sector[p_state->offset-1] & 0x01) << 8;
  offset |= (unsigned short)(p_state->sector[p_state->offset++]);
  //get param1, param2
  param1 = p_state->vars[p_state->sector[p_state->offset++]];
  param2 = p_state->vars[p_state->sector[p_state->offset++]];
  //jump if not greater (signed)
  if( (signed char)param1 <= (signed char)param2 )
    p_state->offset = offset;
}

//-------------------------------------------------------------------------------------------
// JMP_NGTS_C (bit0 is bit8 of offset) <offset> <param> <constant>
// jump if not greater than
// jump to offset if <param> <= <constant>
//-------------------------------------------------------------------------------------------
void am_dev_JMP_NGTS_C( st_am_state * p_state )
{
  unsigned short offset;
  unsigned char param, constant;
  //get offset
  offset = (unsigned short)(p_state->sector[p_state->offset-1] & 0x01) << 8;
  offset |= (unsigned short)(p_state->sector[p_state->offset++]);
  //get param, constant
  param = p_state->vars[p_state->sector[p_state->offset++]];
  constant = p_state->sector[p_state->offset++];
  //jump if not greater (signed)
  if( (signed char)param <= (signed char)constant )
    p_state->offset = offset;
}

//-------------------------------------------------------------------------------------------
// JMP_LTS (bit0 is bit8 of offset) <offset> <param1> <param2>
// jump if less (signed) than
// jump to offset if <param1> < <param2>
//-------------------------------------------------------------------------------------------
void am_dev_JMP_LTS( st_am_state * p_state )
{
  unsigned short offset;
  unsigned char param1, param2;
  //get offset
  offset = (unsigned short)(p_state->sector[p_state->offset-1] & 0x01) << 8;
  offset |= (unsigned short)(p_state->sector[p_state->offset++]);
  //get param1, param2
  param1 = p_state->vars[p_state->sector[p_state->offset++]];
  param2 = p_state->vars[p_state->sector[p_state->offset++]];
  //jump if less (signed)
  if( (signed char)param1 < (signed char)param2 )
    p_state->offset = offset;
}

//-------------------------------------------------------------------------------------------
// JMP_LTS_C (bit0 is bit8 of offset) <offset> <param> <constant>
// jump if less (signed) than
// jump to offset if <param> < <constant>
//-------------------------------------------------------------------------------------------
void am_dev_JMP_LTS_C( st_am_state * p_state )
{
  unsigned short offset;
  unsigned char param, constant;
  //get offset
  offset = (unsigned short)(p_state->sector[p_state->offset-1] & 0x01) << 8;
  offset |= (unsigned short)(p_state->sector[p_state->offset++]);
  //get param, constant
  param = p_state->vars[p_state->sector[p_state->offset++]];
  constant = p_state->sector[p_state->offset++];
  //jump if less (signed)
  if( (signed char)param < (signed char)constant )
    p_state->offset = offset;
}

//-------------------------------------------------------------------------------------------
// JMP_NLTS (bit0 is bit8 of offset) <offset> <param1> <param2>
// jump if not less (signed) than
// jump to offset if <param1> >= <param2>
//-------------------------------------------------------------------------------------------
void am_dev_JMP_NLTS( st_am_state * p_state )
{
  unsigned short offset;
  unsigned char param1, param2;
  //get offset
  offset = (unsigned short)(p_state->sector[p_state->offset-1] & 0x01) << 8;
  offset |= (unsigned short)(p_state->sector[p_state->offset++]);
  //get param1, param2
  param1 = p_state->vars[p_state->sector[p_state->offset++]];
  param2 = p_state->vars[p_state->sector[p_state->offset++]];
  //jump if not less (signed)
  if( (signed char)param1 >= (signed char)param2 )
    p_state->offset = offset;
}

//-------------------------------------------------------------------------------------------
// JMP_NLTS_C (bit0 is bit8 of offset) <offset> <param> <constant>
// jump if not less (signed) than
// jump to offset if <param> >= <constant>
//-------------------------------------------------------------------------------------------
void am_dev_JMP_NLTS_C( st_am_state * p_state )
{
  unsigned short offset;
  unsigned char param, constant;
  //get offset
  offset = (unsigned short)(p_state->sector[p_state->offset-1] & 0x01) << 8;
  offset |= (unsigned short)(p_state->sector[p_state->offset++]);
  //get param, constant
  param = p_state->vars[p_state->sector[p_state->offset++]];
  constant = p_state->sector[p_state->offset++];
  //jump if not less (signed)
  if( (signed char)param >= (signed char)constant )
    p_state->offset = offset;
}

//-------------------------------------------------------------------------------------------
// JMP_AND_Z (bit0 is bit8 of offset) <offset> <param1> <param2>
// jump if bitwise and is zero
// jump to offset if <param1> AND <param2> == 0
//-------------------------------------------------------------------------------------------
void am_dev_JMP_AND_Z( st_am_state * p_state )
{
  unsigned short offset;
  unsigned char param1, param2;
  //get offset
  offset = (unsigned short)(p_state->sector[p_state->offset-1] & 0x01) << 8;
  offset |= (unsigned short)(p_state->sector[p_state->offset++]);
  //get param1, param2
  param1 = p_state->vars[p_state->sector[p_state->offset++]];
  param2 = p_state->vars[p_state->sector[p_state->offset++]];
  //jump if bitwise-and is 0
  if( (param1 & param2) == 0 )
    p_state->offset = offset;
}

//-------------------------------------------------------------------------------------------
// JMP_AND_Z_C (bit0 is bit8 of offset) <offset> <param> <constant>
// jump if bitwise and is zero
// jump to offset if <param> AND <constant> == 0
//-------------------------------------------------------------------------------------------
void am_dev_JMP_AND_Z_C( st_am_state * p_state )
{
  unsigned short offset;
  unsigned char param, constant;
  //get offset
  offset = (unsigned short)(p_state->sector[p_state->offset-1] & 0x01) << 8;
  offset |= (unsigned short)(p_state->sector[p_state->offset++]);
  //get param, constant
  param = p_state->vars[p_state->sector[p_state->offset++]];
  constant = p_state->sector[p_state->offset++];
  //jump if bitwise-and is 0
  if( (param & constant) == 0 )
    p_state->offset = offset;
}

//-------------------------------------------------------------------------------------------
// JMP_AND_NZ (bit0 is bit8 of offset) <offset> <param1> <param2>
// jump if bitwise and is not zero
// jump to offset if <param1> AND <param2> != 0
//-------------------------------------------------------------------------------------------
void am_dev_JMP_AND_NZ( st_am_state * p_state )
{
  unsigned short offset;
  unsigned char param1, param2;
  //get offset
  offset = (unsigned short)(p_state->sector[p_state->offset-1] & 0x01) << 8;
  offset |= (unsigned short)(p_state->sector[p_state->offset++]);
  //get param1, param2
  param1 = p_state->vars[p_state->sector[p_state->offset++]];
  param2 = p_state->vars[p_state->sector[p_state->offset++]];
  //jump if bitwise-and is not 0
  if( (param1 & param2) != 0 )
    p_state->offset = offset;
}

//-------------------------------------------------------------------------------------------
// JMP_AND_NZ_C (bit0 is bit8 of offset) <offset> <param> <constant>
// jump if bitwise and is not zero
// jump to offset if <param> AND <constant> != 0
//-------------------------------------------------------------------------------------------
void am_dev_JMP_AND_NZ_C( st_am_state * p_state )
{
  unsigned short offset;
  unsigned char param, constant;
  //get offset
  offset = (unsigned short)(p_state->sector[p_state->offset-1] & 0x01) << 8;
  offset |= (unsigned short)(p_state->sector[p_state->offset++]);
  //get param, constant
  param = p_state->vars[p_state->sector[p_state->offset++]];
  constant = p_state->sector[p_state->offset++];
  //jump if bitwise-and is not 0
  if( (param & constant) != 0 )
    p_state->offset = offset;
}

//-------------------------------------------------------------------------------------------
// JMP_OR_M (bit0 is bit8 of offset) <offset> <param1> <param2>
// jump if bitwise or is maximum
// jump to offset if <param1> OR <param2> == 0xFF
//-------------------------------------------------------------------------------------------
void am_dev_JMP_OR_M( st_am_state * p_state )
{
  unsigned short offset;
  unsigned char param1, param2;
  //get offset
  offset = (unsigned short)(p_state->sector[p_state->offset-1] & 0x01) << 8;
  offset |= (unsigned short)(p_state->sector[p_state->offset++]);
  //get param1, param2
  param1 = p_state->vars[p_state->sector[p_state->offset++]];
  param2 = p_state->vars[p_state->sector[p_state->offset++]];
  //jump if bitwise-or is maximum
  if( (param1 | param2) == 0xFF )
    p_state->offset = offset;
}

//-------------------------------------------------------------------------------------------
// JMP_OR_M_C (bit0 is bit8 of offset) <offset> <param> <constant>
// jump if bitwise or is maximum
// jump to offset if <param> OR <constant> == 0xFF
//-------------------------------------------------------------------------------------------
void am_dev_JMP_OR_M_C( st_am_state * p_state )
{
  unsigned short offset;
  unsigned char param, constant;
  //get offset
  offset = (unsigned short)(p_state->sector[p_state->offset-1] & 0x01) << 8;
  offset |= (unsigned short)(p_state->sector[p_state->offset++]);
  //get param, constant
  param = p_state->vars[p_state->sector[p_state->offset++]];
  constant = p_state->sector[p_state->offset++];
  //jump if bitwise-or is maximum
  if( (param | constant) == 0xFF )
    p_state->offset = offset;
}

//-------------------------------------------------------------------------------------------
// JMP_OR_NM (bit0 is bit8 of offset) <offset> <param1> <param2>
// jump if bitwise or is not maximum
// jump to offset if <param1> OR <param2> != 0xFF
//-------------------------------------------------------------------------------------------
void am_dev_JMP_OR_NM( st_am_state * p_state )
{
  unsigned short offset;
  unsigned char param1, param2;
  //get offset
  offset = (unsigned short)(p_state->sector[p_state->offset-1] & 0x01) << 8;
  offset |= (unsigned short)(p_state->sector[p_state->offset++]);
  //get param1, param2
  param1 = p_state->vars[p_state->sector[p_state->offset++]];
  param2 = p_state->vars[p_state->sector[p_state->offset++]];
  //jump if bitwise-or is not maximum
  if( (param1 | param2) != 0xFF )
    p_state->offset = offset;
}

//-------------------------------------------------------------------------------------------
// JMP_OR_NM_C (bit0 is bit8 of offset) <offset> <param> <constant>
// jump if bitwise or is not maximum
// jump to offset if <param> OR <constant> != 0xFF
//-------------------------------------------------------------------------------------------
void am_dev_JMP_OR_NM_C( st_am_state * p_state )
{
  unsigned short offset;
  unsigned char param, constant;
  //get offset
  offset = (unsigned short)(p_state->sector[p_state->offset-1] & 0x01) << 8;
  offset |= (unsigned short)(p_state->sector[p_state->offset++]);
  //get param, constant
  param = p_state->vars[p_state->sector[p_state->offset++]];
  constant = p_state->sector[p_state->offset++];
  //jump if bitwise-or is not maximum
  if( (param | constant) != 0xFF )
    p_state->offset = offset;
}

//-------------------------------------------------------------------------------------------
// MOV <source> <dest>
// move
// write contents of source into dest
//-------------------------------------------------------------------------------------------
void am_dev_MOV( st_am_state * p_state )
{
  unsigned char source;
  //get source
  source = p_state->vars[p_state->sector[p_state->offset++]];
  //write to dest
  p_state->vars[p_state->sector[p_state->offset++]] = source;
}

//-------------------------------------------------------------------------------------------
// MOV_C <constant> <dest>
// move constant
// write constant into dest
//-------------------------------------------------------------------------------------------
void am_dev_MOV_C( st_am_state * p_state )
{
  unsigned char constant;
  //get constant
  constant = p_state->sector[p_state->offset++];
  //write to dest
  p_state->vars[p_state->sector[p_state->offset++]] = constant;
}

//-------------------------------------------------------------------------------------------
// MOV2 <source1> <dest1> <source2> <dest2>
// move 2
// write contents of source<x> into dest<x>
//-------------------------------------------------------------------------------------------
void am_dev_MOV2( st_am_state * p_state )
{
  int i;
  unsigned char source;
  //2 times
  for( i = 0; i < 2; i++ )
  {
    //get source
    source = p_state->vars[p_state->sector[p_state->offset++]];
    //write to dest
    p_state->vars[p_state->sector[p_state->offset++]] = source;
  }
}

//-------------------------------------------------------------------------------------------
// MOV2_C <constant1> <dest1> <constant2> <dest2>
// move 2 constants
// write constant<x> into dest<x>
//-------------------------------------------------------------------------------------------
void am_dev_MOV2_C( st_am_state * p_state )
{
  int i;
  unsigned char constant;
  //2 times
  for( i = 0; i < 2; i++ )
  {
    //get constant
    constant = p_state->sector[p_state->offset++];
    //write to dest
    p_state->vars[p_state->sector[p_state->offset++]] = constant;
  }
}

//-------------------------------------------------------------------------------------------
// MOV4 <source1> <dest1> ... <source4> <dest4>
// move 4
// write contents of source<x> into dest<x>
//-------------------------------------------------------------------------------------------
void am_dev_MOV4( st_am_state * p_state )
{
  int i;
  unsigned char source;
  //2 times
  for( i = 0; i < 4; i++ )
  {
    //get source
    source = p_state->vars[p_state->sector[p_state->offset++]];
    //write to dest
    p_state->vars[p_state->sector[p_state->offset++]] = source;
  }
}

//-------------------------------------------------------------------------------------------
// MOV4_C <constant1> <dest1> ... <constant4> <dest4>
// move 4 constants
// write constant<x> into dest<x>
//-------------------------------------------------------------------------------------------
void am_dev_MOV4_C( st_am_state * p_state )
{
  int i;
  unsigned char constant;
  //2 times
  for( i = 0; i < 4; i++ )
  {
    //get constant
    constant = p_state->sector[p_state->offset++];
    //write to dest
    p_state->vars[p_state->sector[p_state->offset++]] = constant;
  }
}

//-------------------------------------------------------------------------------------------
// RD_I <addr> <dest>
// read indirect
// write contents of variable pointed to by addr into dest
//-------------------------------------------------------------------------------------------
void am_dev_RD_I( st_am_state * p_state )
{
  unsigned char addr;
  //get address
  addr = p_state->vars[p_state->sector[p_state->offset++]];
  //write contents of address to dest
  p_state->vars[p_state->sector[p_state->offset++]] = p_state->vars[addr];
}

//-------------------------------------------------------------------------------------------
// WR_I <source> <addr>
// write indirect
// write contents of source into variable pointed to by addr
//-------------------------------------------------------------------------------------------
void am_dev_WR_I( st_am_state * p_state )
{
  unsigned char source, addr;
  //get source, address
  source = p_state->vars[p_state->sector[p_state->offset++]];
  addr = p_state->vars[p_state->sector[p_state->offset++]];
  //write source to place pointed to by address
  p_state->vars[addr] = source;
}

//-------------------------------------------------------------------------------------------
// RD_S <offset_h> <offset_l> <dest>
// read from current sector
// write byte at offset of current sector into dest
//-------------------------------------------------------------------------------------------
void am_dev_RD_S( st_am_state * p_state )
{
  unsigned short offset;
  //get offset
  offset = (unsigned short)(p_state->vars[p_state->sector[p_state->offset++]] & 0x01) << 8;
  offset |= (unsigned short)(p_state->vars[p_state->sector[p_state->offset++]]);
  //write byte at offset of current sector to dest
  p_state->vars[p_state->sector[p_state->offset++]] = p_state->sector[offset];
}

//-------------------------------------------------------------------------------------------
// ADD <source1> <source2> <dest>
// add
// write sum of contents of source1 and contents of source2 into dest
//-------------------------------------------------------------------------------------------
void am_dev_ADD( st_am_state * p_state )
{
  unsigned char source1, source2;
  //get source1, source2
  source1 = p_state->vars[p_state->sector[p_state->offset++]];
  source2 = p_state->vars[p_state->sector[p_state->offset++]];
  //write sum to dest
  p_state->vars[p_state->sector[p_state->offset++]] = source1 + source2;
}

//-------------------------------------------------------------------------------------------
// ADD_C <constant> <source> <dest>
// add constant
// write sum of constant and contents of source into dest
//-------------------------------------------------------------------------------------------
void am_dev_ADD_C( st_am_state * p_state )
{
  unsigned char constant, source;
  //get constant, source
  constant = p_state->sector[p_state->offset++];
  source = p_state->vars[p_state->sector[p_state->offset++]];
  //write sum to dest
  p_state->vars[p_state->sector[p_state->offset++]] = constant + source;
}

//-------------------------------------------------------------------------------------------
// ADDC <source1> <source2> <carry_in> <dest> <carry_out>
// add (with carry)
// write sum of contents of source1 and contents of source2 into dest
// bit 0 of carry_in is also included in the sum
// carry_out is 0x01 if the result is >= 0x100, 0x00 otherwise
//-------------------------------------------------------------------------------------------
void am_dev_ADDC( st_am_state * p_state )
{
  unsigned char source1, source2, carry_in;
  unsigned short sum;
  //get source1, source2, carry_in
  source1 = p_state->vars[p_state->sector[p_state->offset++]];
  source2 = p_state->vars[p_state->sector[p_state->offset++]];
  carry_in = p_state->vars[p_state->sector[p_state->offset++]] & 0x01;
  //calculate sum
  sum = (unsigned short)source1 + (unsigned short)source2 + (unsigned short)carry_in;
  //write sum to dest and carry_out
  p_state->vars[p_state->sector[p_state->offset++]] = (unsigned char)(sum & 0xFF);
  p_state->vars[p_state->sector[p_state->offset++]] = (unsigned char)(sum >> 8);
}

//-------------------------------------------------------------------------------------------
// ADDC_C <constant> <source> <carry_in> <dest> <carry_out>
// add constant (with carry)
// write sum of constant and contents of source into dest
// bit 0 of carry_in is also included in the sum
// carry_out is 0x01 if the result is >= 0x100, 0x00 otherwise
//-------------------------------------------------------------------------------------------
void am_dev_ADDC_C( st_am_state * p_state )
{
  unsigned char constant, source, carry_in;
  unsigned short sum;
  //get constant, source, carry_in
  constant = p_state->sector[p_state->offset++];
  source = p_state->vars[p_state->sector[p_state->offset++]];
  carry_in = p_state->vars[p_state->sector[p_state->offset++]] & 0x01;
  //calculate sum
  sum = (unsigned short)constant + (unsigned short)source + (unsigned short)carry_in;
  //write sum to dest and carry_out
  p_state->vars[p_state->sector[p_state->offset++]] = (unsigned char)(sum & 0xFF);
  p_state->vars[p_state->sector[p_state->offset++]] = (unsigned char)(sum >> 8);
}

//-------------------------------------------------------------------------------------------
// SUB <source1> <source2> <dest>
// subtract
// write difference of contents of source1 minus contents of source2 into dest
//-------------------------------------------------------------------------------------------
void am_dev_SUB( st_am_state * p_state )
{
  unsigned char source1, source2;
  //get source1, source2
  source1 = p_state->vars[p_state->sector[p_state->offset++]];
  source2 = p_state->vars[p_state->sector[p_state->offset++]];
  //write difference to dest
  p_state->vars[p_state->sector[p_state->offset++]] = source1 - source2;
}

//-------------------------------------------------------------------------------------------
// SUB_C <constant> <source> <dest>
// subtract from constant
// write difference of constant minus contents of source into dest
//-------------------------------------------------------------------------------------------
void am_dev_SUB_C( st_am_state * p_state )
{
  unsigned char constant, source;
  //get constant, source
  constant = p_state->sector[p_state->offset++];
  source = p_state->vars[p_state->sector[p_state->offset++]];
  //write difference to dest
  p_state->vars[p_state->sector[p_state->offset++]] = constant - source;
}

//-------------------------------------------------------------------------------------------
// SUBB <source1> <source2> <borrow_in> <dest> <borrow_out>
// subtract (with borrow)
// write difference of contents of source1 minus contents of source2 into dest
// bit 0 of borrow_in is also subtracted
// borrow_out is 0x01 if the result is < 0x00, 0x00 otherwise
//-------------------------------------------------------------------------------------------
void am_dev_SUBB( st_am_state * p_state )
{
  unsigned char source1, source2, borrow_in;
  unsigned short diff;
  //get source1, source2, borrow_in
  source1 = p_state->vars[p_state->sector[p_state->offset++]];
  source2 = p_state->vars[p_state->sector[p_state->offset++]];
  borrow_in = p_state->vars[p_state->sector[p_state->offset++]] & 0x01;
  //calculate difference
  diff = (unsigned short)source1 - (unsigned short)source2 - (unsigned short)borrow_in;
  //write difference to dest and borrow_out
  p_state->vars[p_state->sector[p_state->offset++]] = (unsigned char)(diff & 0xFF);
  p_state->vars[p_state->sector[p_state->offset++]] = (unsigned char)(diff >> 8) & 0x01;
}

//-------------------------------------------------------------------------------------------
// SUBB_C <constant> <source> <borrow_in> <dest> <borrow_out>
// subtract from constant (with borrow)
// write difference of constant minus contents of source into dest
//-------------------------------------------------------------------------------------------
void am_dev_SUBB_C( st_am_state * p_state )
{
  unsigned char constant, source, borrow_in;
  unsigned short diff;
  //get constant, source, borrow_in
  constant = p_state->sector[p_state->offset++];
  source = p_state->vars[p_state->sector[p_state->offset++]];
  borrow_in = p_state->vars[p_state->sector[p_state->offset++]] & 0x01;
  //calculate difference
  diff = (unsigned short)constant - (unsigned short)source - (unsigned short)borrow_in;
  //write difference to dest and borrow_out
  p_state->vars[p_state->sector[p_state->offset++]] = (unsigned char)(diff & 0xFF);
  p_state->vars[p_state->sector[p_state->offset++]] = (unsigned char)(diff >> 8) & 0x01;
}

//-------------------------------------------------------------------------------------------
// MUL <source1> <source2> <dest_h> <dest_l>
// multiply
// write product of contents of source1 and contents of source2 into dest_h and dest_l
//-------------------------------------------------------------------------------------------
void am_dev_MUL( st_am_state * p_state )
{
  unsigned char source1, source2;
  unsigned short prod;
  //get source1, source2
  source1 = p_state->vars[p_state->sector[p_state->offset++]];
  source2 = p_state->vars[p_state->sector[p_state->offset++]];
  //write product to dest
  prod = (unsigned short)source1 * (unsigned short)source2;
  p_state->vars[p_state->sector[p_state->offset++]] = (unsigned char)(prod >> 8);
  p_state->vars[p_state->sector[p_state->offset++]] = (unsigned char)(prod & 0xFF);
}

//-------------------------------------------------------------------------------------------
// MUL_C <constant> <source> <dest_h> <dest_l>
// multiply with constant
// write product of constant and contents of source into dest_h and dest_l
//-------------------------------------------------------------------------------------------
void am_dev_MUL_C( st_am_state * p_state )
{
  unsigned char constant, source;
  unsigned short prod;
  //get constant, source
  constant = p_state->sector[p_state->offset++];
  source = p_state->vars[p_state->sector[p_state->offset++]];
  //write product to dest
  prod = (unsigned short)constant * (unsigned short)source;
  p_state->vars[p_state->sector[p_state->offset++]] = (unsigned char)(prod >> 8);
  p_state->vars[p_state->sector[p_state->offset++]] = (unsigned char)(prod & 0xFF);
}

//-------------------------------------------------------------------------------------------
// MULS <source1> <source2> <dest_h> <dest_l>
// multiply (signed)
// write product of contents of source1 and contents of source2 into dest_h and dest_l
//-------------------------------------------------------------------------------------------
void am_dev_MULS( st_am_state * p_state )
{
  unsigned char source1, source2;
  signed short prod;
  //get source1, source2
  source1 = p_state->vars[p_state->sector[p_state->offset++]];
  source2 = p_state->vars[p_state->sector[p_state->offset++]];
  //write product to dest
  prod = (signed short)source1 * (signed short)source2;
  p_state->vars[p_state->sector[p_state->offset++]] = (unsigned char)(prod >> 8);
  p_state->vars[p_state->sector[p_state->offset++]] = (unsigned char)(prod & 0xFF);
}

//-------------------------------------------------------------------------------------------
// MULS_C <constant> <source> <dest_h> <dest_l>
// multiply with constant (signed)
// write product of constant and contents of source into dest_h and dest_l
//-------------------------------------------------------------------------------------------
void am_dev_MULS_C( st_am_state * p_state )
{
  unsigned char constant, source;
  signed short prod;
  //get constant, source
  constant = p_state->sector[p_state->offset++];
  source = p_state->vars[p_state->sector[p_state->offset++]];
  //write product to dest
  prod = (signed short)constant * (signed short)source;
  p_state->vars[p_state->sector[p_state->offset++]] = (unsigned char)(prod >> 8);
  p_state->vars[p_state->sector[p_state->offset++]] = (unsigned char)(prod & 0xFF);
}

//-------------------------------------------------------------------------------------------
// AND <source1> <source2> <dest>
// bitwise and
// write bitwise and of contents of source1 and contents of source2 into dest
//-------------------------------------------------------------------------------------------
void am_dev_AND( st_am_state * p_state )
{
  unsigned char source1, source2;
  //get source1, source2
  source1 = p_state->vars[p_state->sector[p_state->offset++]];
  source2 = p_state->vars[p_state->sector[p_state->offset++]];
  //write bitwise-and to dest
  p_state->vars[p_state->sector[p_state->offset++]] = source1 & source2;
}

//-------------------------------------------------------------------------------------------
// AND_C <constant> <source> <dest>
// and bitwise with constant
// write bitwise and of constant and contents of source into dest
//-------------------------------------------------------------------------------------------
void am_dev_AND_C( st_am_state * p_state )
{
  unsigned char constant, source;
  //get constant, source
  constant = p_state->sector[p_state->offset++];
  source = p_state->vars[p_state->sector[p_state->offset++]];
  //write bitwise-and to dest
  p_state->vars[p_state->sector[p_state->offset++]] = constant & source;
}

//-------------------------------------------------------------------------------------------
// OR <source1> <source2> <dest>
// bitwise or
// write bitwise or of contents of source1 and contents of source2 into dest
//-------------------------------------------------------------------------------------------
void am_dev_OR( st_am_state * p_state )
{
  unsigned char source1, source2;
  //get source1, source2
  source1 = p_state->vars[p_state->sector[p_state->offset++]];
  source2 = p_state->vars[p_state->sector[p_state->offset++]];
  //write bitwise-or to dest
  p_state->vars[p_state->sector[p_state->offset++]] = source1 | source2;
}

//-------------------------------------------------------------------------------------------
// OR_C <constant> <source> <dest>
// or bitwise with constant
// write bitwise or of constant and contents of source into dest
//-------------------------------------------------------------------------------------------
void am_dev_OR_C( st_am_state * p_state )
{
  unsigned char constant, source;
  //get constant, source
  constant = p_state->sector[p_state->offset++];
  source = p_state->vars[p_state->sector[p_state->offset++]];
  //write bitwise-or to dest
  p_state->vars[p_state->sector[p_state->offset++]] = constant | source;
}

//-------------------------------------------------------------------------------------------
// XOR <source1> <source2> <dest>
// bitwise xor
// write bitwise xor of contents of source1 and contents of source2 into dest
//-------------------------------------------------------------------------------------------
void am_dev_XOR( st_am_state * p_state )
{
  unsigned char source1, source2;
  //get source1, source2
  source1 = p_state->vars[p_state->sector[p_state->offset++]];
  source2 = p_state->vars[p_state->sector[p_state->offset++]];
  //write bitwise-xor to dest
  p_state->vars[p_state->sector[p_state->offset++]] = source1 ^ source2;
}

//-------------------------------------------------------------------------------------------
// XOR_C <constant> <source> <dest>
// xor bitwise with constant
// write bitwise xor of constant and contents of source into dest
//-------------------------------------------------------------------------------------------
void am_dev_XOR_C( st_am_state * p_state )
{
  unsigned char constant, source;
  //get constant, source
  constant = p_state->sector[p_state->offset++];
  source = p_state->vars[p_state->sector[p_state->offset++]];
  //write bitwise-xor to dest
  p_state->vars[p_state->sector[p_state->offset++]] = constant ^ source;
}

//-------------------------------------------------------------------------------------------
// SHL <bits> <source> <dest>
// shift left
// shift left contents of source <bits> bits and write result into dest
//-------------------------------------------------------------------------------------------
void am_dev_SHL( st_am_state * p_state )
{
  unsigned char bits, source;
  //get bits, source
  bits = p_state->vars[p_state->sector[p_state->offset++]];
  source = p_state->vars[p_state->sector[p_state->offset++]];
  //write shifted value to dest
  p_state->vars[p_state->sector[p_state->offset++]] = source << bits;
}

//-------------------------------------------------------------------------------------------
// SHL_C <constant> <source> <dest>
// shift left constant number of bits
// shift left contents of source <constant> bits and write result into dest
//-------------------------------------------------------------------------------------------
void am_dev_SHL_C( st_am_state * p_state )
{
  unsigned char constant, source;
  //get constant, source
  constant = p_state->sector[p_state->offset++];
  source = p_state->vars[p_state->sector[p_state->offset++]];
  //write shifted value to dest
  p_state->vars[p_state->sector[p_state->offset++]] = source << constant;
}

//-------------------------------------------------------------------------------------------
// SHR <bits> <source> <dest>
// shift right
// shift right contents of source <bits> bits and write result into dest
//-------------------------------------------------------------------------------------------
void am_dev_SHR( st_am_state * p_state )
{
  unsigned char bits, source;
  //get bits, source
  bits = p_state->vars[p_state->sector[p_state->offset++]];
  source = p_state->vars[p_state->sector[p_state->offset++]];
  //write shifted value to dest
  p_state->vars[p_state->sector[p_state->offset++]] = source >> bits;
}

//-------------------------------------------------------------------------------------------
// SHR_C <constant> <source> <dest>
// shift right constant number of bits
// shift right contents of source <constant> bits and write result into dest
//-------------------------------------------------------------------------------------------
void am_dev_SHR_C( st_am_state * p_state )
{
  unsigned char constant, source;
  //get constant, source
  constant = p_state->sector[p_state->offset++];
  source = p_state->vars[p_state->sector[p_state->offset++]];
  //write shifted value to dest
  p_state->vars[p_state->sector[p_state->offset++]] = source >> constant;
}

//-------------------------------------------------------------------------------------------
// ROL <bits> <source> <dest>
// rotate left
// rotate left contents of source <bits> bits and write result into dest
//-------------------------------------------------------------------------------------------
void am_dev_ROL( st_am_state * p_state )
{
  unsigned char bits, source, rotated;
  //get bits, source
  bits = p_state->vars[p_state->sector[p_state->offset++]];
  source = p_state->vars[p_state->sector[p_state->offset++]];
  //rotate value
  rotated = (source << (bits & 7)) | (source >> (8 - (bits & 7)));
  //write rotated value to dest
  p_state->vars[p_state->sector[p_state->offset++]] = rotated;
}

//-------------------------------------------------------------------------------------------
// ROL_C <constant> <source> <dest>
// rotate left constant number of bits
// rotate left contents of source <constant> bits and write result into dest
//-------------------------------------------------------------------------------------------
void am_dev_ROL_C( st_am_state * p_state )
{
  unsigned char constant, source, rotated;
  //get constant, source
  constant = p_state->sector[p_state->offset++];
  source = p_state->vars[p_state->sector[p_state->offset++]];
  //rotate value
  rotated = (source << (constant & 7)) | (source >> (8 - (constant & 7)));
  //write rotated value to dest
  p_state->vars[p_state->sector[p_state->offset++]] = rotated;
}

//-------------------------------------------------------------------------------------------
// ROR <bits> <source> <dest>
// rotate right
// rotate right contents of source <bits> bits and write result into dest
//-------------------------------------------------------------------------------------------
void am_dev_ROR( st_am_state * p_state )
{
  unsigned char bits, source, rotated;
  //get bits, source
  bits = p_state->vars[p_state->sector[p_state->offset++]];
  source = p_state->vars[p_state->sector[p_state->offset++]];
  //rotate value
  rotated = (source >> (bits & 7)) | (source << (8 - (bits & 7)));
  //write rotated value to dest
  p_state->vars[p_state->sector[p_state->offset++]] = rotated;
}

//-------------------------------------------------------------------------------------------
// ROR_C <constant> <source> <dest>
// rotate right constant number of bits
// rotate right contents of source <constant> bits and write result into dest
//-------------------------------------------------------------------------------------------
void am_dev_ROR_C( st_am_state * p_state )
{
  unsigned char constant, source, rotated;
  //get constant, source
  constant = p_state->sector[p_state->offset++];
  source = p_state->vars[p_state->sector[p_state->offset++]];
  //rotate value
  rotated = (source >> (constant & 7)) | (source << (8 - (constant & 7)));
  //write rotated value to dest
  p_state->vars[p_state->sector[p_state->offset++]] = rotated;
}

//-------------------------------------------------------------------------------------------
// RD_PIX <var_x> <var_y> <dest>
// read pixel
// write the value of the pixel at position var_x, var_y into dest
//-------------------------------------------------------------------------------------------
void am_dev_RD_PIX( st_am_state * p_state )
{
  unsigned char var_x, var_y, value;
  //get var_x, var_y
  var_x = p_state->vars[p_state->sector[p_state->offset++]];
  var_y = p_state->vars[p_state->sector[p_state->offset++]];
  //get pixel
  if( var_x < 26 && var_y < 20 )
    value = p_state->pixels[var_y][var_x];
  else
    value = 0;
  //write value to dest
  p_state->vars[p_state->sector[p_state->offset++]] = value;
}

//-------------------------------------------------------------------------------------------
// RD_PIX_C <con_x> <con_y> <dest>
// read constant pixel
// write the value of the pixel at position con_x, con_y into dest
//-------------------------------------------------------------------------------------------
void am_dev_RD_PIX_C( st_am_state * p_state )
{
  unsigned char con_x, con_y, value;
  //get con_x, con_y
  con_x = p_state->sector[p_state->offset++];
  con_y = p_state->sector[p_state->offset++];
  //get pixel
  if( con_x < 26 && con_y < 20 )
    value = p_state->pixels[con_y][con_x];
  else
    value = 0;
  //write value to dest
  p_state->vars[p_state->sector[p_state->offset++]] = value;
}

//-------------------------------------------------------------------------------------------
// RD_OVP <var_x> <var_y> <dest>
// read overlay pixel
// write the value of the overlay pixel at position var_x, var_y into dest
//-------------------------------------------------------------------------------------------
void am_dev_RD_OVP( st_am_state * p_state )
{
  unsigned char var_x, var_y, value;
  //get var_x, var_y
  var_x = p_state->vars[p_state->sector[p_state->offset++]];
  var_y = p_state->vars[p_state->sector[p_state->offset++]];
  //get pixel
  if( var_x < 26 && var_y < 20 )
    value = p_state->overlay[var_y][var_x];
  else
    value = 0;
  //write value to dest
  p_state->vars[p_state->sector[p_state->offset++]] = value;
}

//-------------------------------------------------------------------------------------------
// RD_OVP_C <con_x> <con_y> <dest>
// read constant overlay pixel
// write the value of the overlay pixel at position con_x, con_y into dest
//-------------------------------------------------------------------------------------------
void am_dev_RD_OVP_C( st_am_state * p_state )
{
  unsigned char con_x, con_y, value;
  //get con_x, con_y
  con_x = p_state->sector[p_state->offset++];
  con_y = p_state->sector[p_state->offset++];
  //get pixel
  if( con_x < 26 && con_y < 20 )
    value = p_state->overlay[con_y][con_x];
  else
    value = 0;
  //write value to dest
  p_state->vars[p_state->sector[p_state->offset++]] = value;
}

//-------------------------------------------------------------------------------------------
// WR_PIX <var_x> <var_y> <source>
// write pixel
// write the value of source into the pixel at position var_x, var_y
//-------------------------------------------------------------------------------------------
void am_dev_WR_PIX( st_am_state * p_state )
{
  unsigned char var_x, var_y, value;
  //get var_x, var_y, value
  var_x = p_state->vars[p_state->sector[p_state->offset++]];
  var_y = p_state->vars[p_state->sector[p_state->offset++]];
  value = p_state->vars[p_state->sector[p_state->offset++]];
  //get pixel
  if( var_x < 26 && var_y < 20 )
    p_state->pixels[var_y][var_x] = value & 0x0F;
}

//-------------------------------------------------------------------------------------------
// WR_PIX_C <con_x> <con_y> <source>
// write constant pixel
// write the value of source into the pixel at position con_x, con_y
//-------------------------------------------------------------------------------------------
void am_dev_WR_PIX_C( st_am_state * p_state )
{
  unsigned char con_x, con_y, value;
  //get var_x, var_y, value
  con_x = p_state->sector[p_state->offset++];
  con_y = p_state->sector[p_state->offset++];
  value = p_state->vars[p_state->sector[p_state->offset++]];
  //get pixel
  if( con_x < 26 && con_y < 20 )
    p_state->pixels[con_y][con_x] = value & 0x0F;
}

//-------------------------------------------------------------------------------------------
// WR_OVP <var_x> <var_y> <source>
// write overlay pixel
// write the value of source into the overlay pixel at position var_x, var_y
//-------------------------------------------------------------------------------------------
void am_dev_WR_OVP( st_am_state * p_state )
{
  unsigned char var_x, var_y, value;
  //get var_x, var_y, value
  var_x = p_state->vars[p_state->sector[p_state->offset++]];
  var_y = p_state->vars[p_state->sector[p_state->offset++]];
  value = p_state->vars[p_state->sector[p_state->offset++]];
  //get pixel
  if( var_x < 26 && var_y < 20 )
    p_state->overlay[var_y][var_x] = value & 0x0F;
}

//-------------------------------------------------------------------------------------------
// WR_OVP_C <con_x> <con_y> <source>
// write constant overlay pixel
// write the value of source into the overlay pixel at position con_x, con_y
//-------------------------------------------------------------------------------------------
void am_dev_WR_OVP_C( st_am_state * p_state )
{
  unsigned char con_x, con_y, value;
  //get var_x, var_y, value
  con_x = p_state->sector[p_state->offset++];
  con_y = p_state->sector[p_state->offset++];
  value = p_state->vars[p_state->sector[p_state->offset++]];
  //get pixel
  if( con_x < 26 && con_y < 20 )
    p_state->overlay[con_y][con_x] = value & 0x0F;
}

//-------------------------------------------------------------------------------------------
// WR_PIX_CO <var_x> <var_y> <constant>
// set pixel to color
// write const_source into the pixel at position var_x, var_y
//-------------------------------------------------------------------------------------------
void am_dev_WR_PIX_CO( st_am_state * p_state )
{
  unsigned char var_x, var_y, value;
  //get var_x, var_y, value
  var_x = p_state->vars[p_state->sector[p_state->offset++]];
  var_y = p_state->vars[p_state->sector[p_state->offset++]];
  value = p_state->sector[p_state->offset++];
  //get pixel
  if( var_x < 26 && var_y < 20 )
    p_state->pixels[var_y][var_x] = value & 0x0F;
}

//-------------------------------------------------------------------------------------------
// WR_PIX_CO_C <con_x> <con_y> <constant>
// set constant pixel to color
// write constant into the pixel at position con_x, con_y
//-------------------------------------------------------------------------------------------
void am_dev_WR_PIX_CO_C( st_am_state * p_state )
{
  unsigned char con_x, con_y, value;
  //get var_x, var_y, value
  con_x = p_state->sector[p_state->offset++];
  con_y = p_state->sector[p_state->offset++];
  value = p_state->sector[p_state->offset++];
  //get pixel
  if( con_x < 26 && con_y < 20 )
    p_state->pixels[con_y][con_x] = value & 0x0F;
}

//-------------------------------------------------------------------------------------------
// WR_OVP_CO <var_x> <var_y> <constant>
// set overlay pixel to color
// write constant into the overlay pixel at position var_x, var_y
//-------------------------------------------------------------------------------------------
void am_dev_WR_OVP_CO( st_am_state * p_state )
{
  unsigned char var_x, var_y, value;
  //get var_x, var_y, value
  var_x = p_state->vars[p_state->sector[p_state->offset++]];
  var_y = p_state->vars[p_state->sector[p_state->offset++]];
  value = p_state->sector[p_state->offset++];
  //get pixel
  if( var_x < 26 && var_y < 20 )
    p_state->overlay[var_y][var_x] = value & 0x0F;
}

//-------------------------------------------------------------------------------------------
// WR_OVP_CO_C <con_x> <con_y> <constant>
// set constant overlay pixel to color
// write constant into the overlay pixel at position con_x, con_y
//-------------------------------------------------------------------------------------------
void am_dev_WR_OVP_CO_C( st_am_state * p_state )
{
  unsigned char con_x, con_y, value;
  //get var_x, var_y, value
  con_x = p_state->sector[p_state->offset++];
  con_y = p_state->sector[p_state->offset++];
  value = p_state->sector[p_state->offset++];
  //get pixel
  if( con_x < 26 && con_y < 20 )
    p_state->overlay[con_y][con_x] = value & 0x0F;
}

//-------------------------------------------------------------------------------------------
// WR_PIX_LINE <var_y> <pixel_0_1> <pixel_2_3> ... <pixel_24_25>
// write pixel line
// two pixels are packed in one byte - e.g. pixel_0_1=0xF0 turns on pixel 0 and off pixel 1
//-------------------------------------------------------------------------------------------
void am_dev_WR_PIX_LINE( st_am_state * p_state )
{
  unsigned char var_y, i, x, value;
  //get var_y
  var_y = p_state->vars[p_state->sector[p_state->offset++]];
  if( var_y >= 20 )
  {
    p_state->offset += 13;
    return;
  }
  //get pixels
  x = 0;
  for( i = 0; i < 13; i++ )
  {
    value = p_state->sector[p_state->offset++];
    p_state->pixels[var_y][x++] = value >> 4;
    p_state->pixels[var_y][x++] = value & 0x0F;
  }
}

//-------------------------------------------------------------------------------------------
// WR_PIX_LINE_C <con_y> <pixel_0_1> <pixel_2_3> ... <pixel_24_25>
// write constant pixel line
// two pixels are packed in one byte - e.g. pixel_0_1=0xF0 turns on pixel 0 and off pixel 1
//-------------------------------------------------------------------------------------------
void am_dev_WR_PIX_LINE_C( st_am_state * p_state )
{
  unsigned char con_y, i, x, value;
  //get con_y
  con_y = p_state->sector[p_state->offset++];
  if( con_y >= 20 )
  {
    p_state->offset += 13;
    return;
  }
  //get pixels
  x = 0;
  for( i = 0; i < 13; i++ )
  {
    value = p_state->sector[p_state->offset++];
    p_state->pixels[con_y][x++] = value >> 4;
    p_state->pixels[con_y][x++] = value & 0x0F;
  }
}

//-------------------------------------------------------------------------------------------
// WR_OVP_LINE <var_y> <ov-pixel_0_1> <ov-pixel_2_3> ... <ov-pixel_24_25>
// write overlay pixel line
// two overlay pixels are packed in one byte - e.g. overlay pixel_0_1=0xF0 turns on overlay pixel 0 and off overlay pixel 1
//-------------------------------------------------------------------------------------------
void am_dev_WR_OVP_LINE( st_am_state * p_state )
{
  unsigned char var_y, i, x, value;
  //get var_y
  var_y = p_state->vars[p_state->sector[p_state->offset++]];
  if( var_y >= 20 )
  {
    p_state->offset += 13;
    return;
  }
  //get pixels
  x = 0;
  for( i = 0; i < 13; i++ )
  {
    value = p_state->sector[p_state->offset++];
    p_state->overlay[var_y][x++] = value >> 4;
    p_state->overlay[var_y][x++] = value & 0x0F;
  }
}

//-------------------------------------------------------------------------------------------
// WR_OVP_LINE_C <con_y> <ov-pixel_0_1> <ov-pixel_2_3> ... <ov-pixel_24_25>
// write constant overlay pixel line
// two overlay pixels are packed in one byte - e.g. overlay pixel_0_1=0xF0 turns on overlay pixel 0 and off overlay pixel 1
//-------------------------------------------------------------------------------------------
void am_dev_WR_OVP_LINE_C( st_am_state * p_state )
{
  unsigned char con_y, i, x, value;
  //get con_y
  con_y = p_state->sector[p_state->offset++];
  if( con_y >= 20 )
  {
    p_state->offset += 13;
    return;
  }
  //get pixels
  x = 0;
  for( i = 0; i < 13; i++ )
  {
    value = p_state->sector[p_state->offset++];
    p_state->overlay[con_y][x++] = value >> 4;
    p_state->overlay[con_y][x++] = value & 0x0F;
  }
}

//-------------------------------------------------------------------------------------------
// WR_PIX_QUART <var_quart> <pixel_l0_0_1> <pixel_l0_2_3> ... <pixel_l4_24_25>
// write pixel quarter
// two pixels are packed in one byte
// - e.g. pixel_l0_0_1=0xF0 turns on line 0, pixel 0 and off pixel 1 (lines rel. to quarter)
//-------------------------------------------------------------------------------------------
void am_dev_WR_PIX_QUART( st_am_state * p_state )
{
  unsigned char var_quart, i, x, y, value;
  //get var_quart
  var_quart = p_state->vars[p_state->sector[p_state->offset++]];
  if( var_quart >= 4 )
  {
    p_state->offset += 65;
    return;
  }
  //get pixels
  x = 0;
  y = var_quart * 5;
  for( i = 0; i < 65; i++ )
  {
    value = p_state->sector[p_state->offset++];
    p_state->pixels[y][x++] = value >> 4;
    p_state->pixels[y][x++] = value & 0x0F;
    if( x >= 26 )
      x = 0, y++;
  }
}

//-------------------------------------------------------------------------------------------
// WR_PIX_QUART_C <con_quart> <pixel_l0_0_1> <pixel_l0_2_3> ... <pixel_l4_24_25>
// write constant pixel quarter
// two pixels are packed in one byte
// - e.g. pixel_l0_0_1=0xF0 turns on line 0, pixel 0 and off pixel 1 (lines rel. to quarter)
//-------------------------------------------------------------------------------------------
void am_dev_WR_PIX_QUART_C( st_am_state * p_state )
{
  unsigned char con_quart, i, x, y, value;
  //get var_quart
  con_quart = p_state->sector[p_state->offset++];
  if( con_quart >= 4 )
  {
    p_state->offset += 65;
    return;
  }
  //get pixels
  x = 0;
  y = con_quart * 5;
  for( i = 0; i < 65; i++ )
  {
    value = p_state->sector[p_state->offset++];
    p_state->pixels[y][x++] = value >> 4;
    p_state->pixels[y][x++] = value & 0x0F;
    if( x >= 26 )
      x = 0, y++;
  }
}

//-------------------------------------------------------------------------------------------
// WR_OVP_QUART <var_quart> <ov-pixel_l0_0_1> <ov-pixel_l0_2_3> ... <ov-pixel_l4_24_25>
// write overlay pixel quarter
// two overlay pixels are packed in one byte
// - e.g. overlay pixel_l0_0_1=0xF0 turns on line 0, overlay pixel 0 and off overlay pixel 1 (lines rel. to quarter)
//-------------------------------------------------------------------------------------------
void am_dev_WR_OVP_QUART( st_am_state * p_state )
{
  unsigned char var_quart, i, x, y, value;
  //get var_quart
  var_quart = p_state->vars[p_state->sector[p_state->offset++]];
  if( var_quart >= 4 )
  {
    p_state->offset += 65;
    return;
  }
  //get pixels
  x = 0;
  y = var_quart * 5;
  for( i = 0; i < 65; i++ )
  {
    value = p_state->sector[p_state->offset++];
    p_state->overlay[y][x++] = value >> 4;
    p_state->overlay[y][x++] = value & 0x0F;
    if( x >= 26 )
      x = 0, y++;
  }
}

//-------------------------------------------------------------------------------------------
// WR_OVP_QUART_C <con_quart> <ov-pixel_l0_0_1> <ov-pixel_l0_2_3> ... <ov-pixel_l4_24_25>
// write constant overlay pixel quarter
// two overlay pixels are packed in one byte
// - e.g. overlay pixel_l0_0_1=0xF0 turns on line 0, overlay pixel 0 and off overlay pixel 1
//-------------------------------------------------------------------------------------------
void am_dev_WR_OVP_QUART_C( st_am_state * p_state )
{
  unsigned char con_quart, i, x, y, value;
  //get var_quart
  con_quart = p_state->sector[p_state->offset++];
  if( con_quart >= 4 )
  {
    p_state->offset += 65;
    return;
  }
  //get pixels
  x = 0;
  y = con_quart * 5;
  for( i = 0; i < 65; i++ )
  {
    value = p_state->sector[p_state->offset++];
    p_state->overlay[y][x++] = value >> 4;
    p_state->overlay[y][x++] = value & 0x0F;
    if( x >= 26 )
      x = 0, y++;
  }
}

//-------------------------------------------------------------------------------------------
// WR_PIX_CLR
// clear pixels
//-------------------------------------------------------------------------------------------
void am_dev_WR_PIX_CLR( st_am_state * p_state )
{
  unsigned char x, y;
  for( y = 0; y < 20; y++ )
    for( x = 0; x < 26; x++ )
      p_state->pixels[y][x] = 0x00;
}

//-------------------------------------------------------------------------------------------
// WR_OVP_CLR
// clear overlay pixels
//-------------------------------------------------------------------------------------------
void am_dev_WR_OVP_CLR( st_am_state * p_state )
{
  unsigned char x, y;
  for( y = 0; y < 20; y++ )
    for( x = 0; x < 26; x++ )
      p_state->overlay[y][x] = 0x00;
}

//-------------------------------------------------------------------------------------------
// WR_PIX_SET
// set pixels
//-------------------------------------------------------------------------------------------
void am_dev_WR_PIX_SET( st_am_state * p_state )
{
  unsigned char x, y;
  for( y = 0; y < 20; y++ )
    for( x = 0; x < 26; x++ )
      p_state->pixels[y][x] = 0x0F;
}

//-------------------------------------------------------------------------------------------
// WR_OVP_SET
// set overlay pixels
//-------------------------------------------------------------------------------------------
void am_dev_WR_OVP_SET( st_am_state * p_state )
{
  unsigned char x, y;
  for( y = 0; y < 20; y++ )
    for( x = 0; x < 26; x++ )
      p_state->overlay[y][x] = 0x0F;
}

//-------------------------------------------------------------------------------------------
// WR_PIX_BW <bw0> ... <bw64>
// write frame of black/white pixels
// bw0,7 = pixel 0,0 / bw0,6 = pixel 1,0 / ... / bw64,0 = pixel 25,19
//-------------------------------------------------------------------------------------------
void am_dev_WR_PIX_BW( st_am_state * p_state )
{
  unsigned char i, j, x, y, value;
  x = 0;
  y = 0;
  //65 bytes
  for( i = 0; i < 65; i++ )
  {
    //get next byte
    value = p_state->sector[p_state->offset++];
    for( j = 0; j < 8; j++ )
    {
      //write pixel
      if( (value & 0x80) != 0 )
        p_state->pixels[y][x] = 0x0F;
      else
        p_state->pixels[y][x] = 0x00;
      //next pixel
      value <<= 1;
      x++;
      if( x >= 26 )
        x = 0, y++;
    }
  }
}

//-------------------------------------------------------------------------------------------
// WR_OVP_BW <bw0> ... <bw64>
// write frame of black/white overlay pixels
// bw0,7 = pixel 0,0 / bw0,6 = pixel 1,0 / ... / bw64,0 = pixel 25,19
//-------------------------------------------------------------------------------------------
void am_dev_WR_OVP_BW( st_am_state * p_state )
{
  unsigned char i, j, x, y, value;
  x = 0;
  y = 0;
  //65 bytes
  for( i = 0; i < 65; i++ )
  {
    //get next byte
    value = p_state->sector[p_state->offset++];
    for( j = 0; j < 8; j++ )
    {
      //write pixel
      if( (value & 0x80) != 0 )
        p_state->overlay[y][x] = 0x0F;
      else
        p_state->overlay[y][x] = 0x00;
      //next pixel
      value <<= 1;
      x++;
      if( x >= 26 )
        x = 0, y++;
    }
  }
}

//-------------------------------------------------------------------------------------------
// WR_PIX_SEL <source> <sel0> ... <sel64>
// set selected pixels of frame to source
// - sel selects pixels
// - pixels with sel=1 are set to source
// sel0,7 = pixel 0,0 / sel0,6 = pixel 1,0 / ... / sel64,0 = pixel 25,19
//-------------------------------------------------------------------------------------------
void am_dev_WR_PIX_SEL( st_am_state * p_state )
{
  unsigned char source, i, j, x, y, value;
  x = 0;
  y = 0;
  //get source
  source = p_state->vars[p_state->sector[p_state->offset++]] & 0x0F;
  //65 bytes
  for( i = 0; i < 65; i++ )
  {
    //get next byte
    value = p_state->sector[p_state->offset++];
    for( j = 0; j < 8; j++ )
    {
      //write pixel
      if( (value & 0x80) != 0 )
        p_state->pixels[y][x] = source;
      //next pixel
      value <<= 1;
      x++;
      if( x >= 26 )
        x = 0, y++;
    }
  }
}

//-------------------------------------------------------------------------------------------
// WR_PIX_SEL_C <constant> <sel0> ... <sel64>
// set selected pixels of frame to constant
// - sel selects pixels
// - pixels with sel=1 are set to constant
// sel0,7 = pixel 0,0 / sel0,6 = pixel 1,0 / ... / sel64,0 = pixel 25,19
//-------------------------------------------------------------------------------------------
void am_dev_WR_PIX_SEL_C( st_am_state * p_state )
{
  unsigned char constant, i, j, x, y, value;
  x = 0;
  y = 0;
  //get constant
  constant = p_state->sector[p_state->offset++] & 0x0F;
  //65 bytes
  for( i = 0; i < 65; i++ )
  {
    //get next byte
    value = p_state->sector[p_state->offset++];
    for( j = 0; j < 8; j++ )
    {
      //write pixel
      if( (value & 0x80) != 0 )
        p_state->pixels[y][x] = constant;
      //next pixel
      value <<= 1;
      x++;
      if( x >= 26 )
        x = 0, y++;
    }
  }
}

//-------------------------------------------------------------------------------------------
// WR_OVP_SEL <source> <sel0> ... <sel64>
// set selected overlay pixels of frame to source
// - sel selects pixels
// - pixels with sel=1 are set to source
// sel0,7 = pixel 0,0 / sel0,6 = pixel 1,0 / ... / sel64,0 = pixel 25,19
//-------------------------------------------------------------------------------------------
void am_dev_WR_OVP_SEL( st_am_state * p_state )
{
  unsigned char source, i, j, x, y, value;
  x = 0;
  y = 0;
  //get source
  source = p_state->vars[p_state->sector[p_state->offset++]] & 0x0F;
  //65 bytes
  for( i = 0; i < 65; i++ )
  {
    //get next byte
    value = p_state->sector[p_state->offset++];
    for( j = 0; j < 8; j++ )
    {
      //write pixel
      if( (value & 0x80) != 0 )
        p_state->overlay[y][x] = source;
      //next pixel
      value <<= 1;
      x++;
      if( x >= 26 )
        x = 0, y++;
    }
  }
}

//-------------------------------------------------------------------------------------------
// WR_OVP_SEL_C <constant> <sel0> ... <sel64>
// set selected overlay pixels of frame to constant
// - sel selects pixels
// - pixels with sel=1 are set to constant
// sel0,7 = pixel 0,0 / sel0,6 = pixel 1,0 / ... / sel64,0 = pixel 25,19
//-------------------------------------------------------------------------------------------
void am_dev_WR_OVP_SEL_C( st_am_state * p_state )
{
  unsigned char constant, i, j, x, y, value;
  x = 0;
  y = 0;
  //get constant
  constant = p_state->sector[p_state->offset++] & 0x0F;
  //65 bytes
  for( i = 0; i < 65; i++ )
  {
    //get next byte
    value = p_state->sector[p_state->offset++];
    for( j = 0; j < 8; j++ )
    {
      //write pixel
      if( (value & 0x80) != 0 )
        p_state->overlay[y][x] = constant;
      //next pixel
      value <<= 1;
      x++;
      if( x >= 26 )
        x = 0, y++;
    }
  }
}

//-------------------------------------------------------------------------------------------
// CP_BLK <flags> <src_x> <src_y> <dest_x> <dest_y> <size_x> <size_y>
// copy a block of pixels
// flags,7: 0=copy from pixel buffer, 1=copy from overlay buffer
// flags,6: 0=copy to pixel buffer, 1=copy to overlay buffer
// flags,5: 0=src_x is a variable, 1=src_x is a constant
// flags,4: 0=src_y is a variable, 1=src_y is a constant
// flags,3: 0=dest_x is a variable, 1=dest_x is a constant
// flags,2: 0=dest_y is a variable, 1=dest_y is a constant
// flags,1: 0=size_x is a variable, 1=size_x is a constant
// flags,0: 0=size_y is a variable, 1=size_y is a constant
// if the two sections overlap, the result is undefined
//-------------------------------------------------------------------------------------------
void am_dev_CP_BLK( st_am_state * p_state )
{
  unsigned char (* src)[26], (* dest)[26];
  unsigned char flags, src_x, src_y, dest_x, dest_y, size_x, size_y, x, y;
  //get parameters
  flags = p_state->sector[p_state->offset++];
  if( (flags & 0x80) == 0 )
    src = p_state->pixels;
  else
    src = p_state->overlay;
  if( (flags & 0x40) == 0 )
    dest = p_state->pixels;
  else
    dest = p_state->overlay;
  src_x = p_state->sector[p_state->offset++];
  if( (flags & 0x20) == 0 )
    src_x = p_state->vars[src_x];
  src_y = p_state->sector[p_state->offset++];
  if( (flags & 0x10) == 0 )
    src_y = p_state->vars[src_y];
  dest_x = p_state->sector[p_state->offset++];
  if( (flags & 0x08) == 0 )
    dest_x = p_state->vars[dest_x];
  dest_y = p_state->sector[p_state->offset++];
  if( (flags & 0x04) == 0 )
    dest_y = p_state->vars[dest_y];
  size_x = p_state->sector[p_state->offset++];
  if( (flags & 0x02) == 0 )
    size_x = p_state->vars[size_x];
  size_y = p_state->sector[p_state->offset++];
  if( (flags & 0x01) == 0 )
    size_y = p_state->vars[size_y];
  //check parameters
  if( src_x >= 26 || src_y >= 20 || dest_x >= 26 || dest_y >= 20 || size_x > 26 || size_y > 20 )
    return;
  if( src_x + size_x > 26 || src_y + size_y > 20 || dest_x + size_x > 26 || dest_y + size_y > 20 )
    return;
  //copy
  for( y = 0; y < size_y; y++ )
    for( x = 0; x < size_x; x++ )
      dest[dest_y + y][dest_x + x] = src[src_y + y][src_x + x];
}

//-------------------------------------------------------------------------------------------
// CP_BLK_TR <flags> <transp> <src_x> <src_y> <dest_x> <dest_y> <size_x> <size_y>
// copy a block of pixels with a transparent color
//  - if the variable transp is not in the range 0x00..0x0F no color is transparent
// flags,7: 0=copy from pixel buffer, 1=copy from overlay buffer
// flags,6: 0=copy to pixel buffer, 1=copy to overlay buffer
// flags,5: 0=src_x is a variable, 1=src_x is a constant
// flags,4: 0=src_y is a variable, 1=src_y is a constant
// flags,3: 0=dest_x is a variable, 1=dest_x is a constant
// flags,2: 0=dest_y is a variable, 1=dest_y is a constant
// flags,1: 0=size_x is a variable, 1=size_x is a constant
// flags,0: 0=size_y is a variable, 1=size_y is a constant
// if the two sections overlap, the result is undefined
//-------------------------------------------------------------------------------------------
void am_dev_CP_BLK_TR( st_am_state * p_state )
{
  unsigned char (* src)[26], (* dest)[26];
  unsigned char flags, transp, src_x, src_y, dest_x, dest_y, size_x, size_y, x, y;
  //get parameters
  flags = p_state->sector[p_state->offset++];
  if( (flags & 0x80) == 0 )
    src = p_state->pixels;
  else
    src = p_state->overlay;
  if( (flags & 0x40) == 0 )
    dest = p_state->pixels;
  else
    dest = p_state->overlay;
  transp = p_state->vars[p_state->sector[p_state->offset++]];
  src_x = p_state->sector[p_state->offset++];
  if( (flags & 0x20) == 0 )
    src_x = p_state->vars[src_x];
  src_y = p_state->sector[p_state->offset++];
  if( (flags & 0x10) == 0 )
    src_y = p_state->vars[src_y];
  dest_x = p_state->sector[p_state->offset++];
  if( (flags & 0x08) == 0 )
    dest_x = p_state->vars[dest_x];
  dest_y = p_state->sector[p_state->offset++];
  if( (flags & 0x04) == 0 )
    dest_y = p_state->vars[dest_y];
  size_x = p_state->sector[p_state->offset++];
  if( (flags & 0x02) == 0 )
    size_x = p_state->vars[size_x];
  size_y = p_state->sector[p_state->offset++];
  if( (flags & 0x01) == 0 )
    size_y = p_state->vars[size_y];
  //check parameters
  if( src_x >= 26 || src_y >= 20 || dest_x >= 26 || dest_y >= 20 || size_x > 26 || size_y > 20 )
    return;
  if( src_x + size_x > 26 || src_y + size_y > 20 || dest_x + size_x > 26 || dest_y + size_y > 20 )
    return;
  //copy with transparent color
  for( y = 0; y < size_y; y++ )
    for( x = 0; x < size_x; x++ )
      if( src[src_y + y][src_x + x] != transp )
        dest[dest_y + y][dest_x + x] = src[src_y + y][src_x + x];
}

//-------------------------------------------------------------------------------------------
// CP_BLK_TR_C <flags> <transp> <src_x> <src_y> <dest_x> <dest_y> <size_x> <size_y>
// copy a block of pixels with a constant transparent color
//  - if the constant transp is not in the range 0x00..0x0F no color is transparent
// flags,7: 0=copy from pixel buffer, 1=copy from overlay buffer
// flags,6: 0=copy to pixel buffer, 1=copy to overlay buffer
// flags,5: 0=src_x is a variable, 1=src_x is a constant
// flags,4: 0=src_y is a variable, 1=src_y is a constant
// flags,3: 0=dest_x is a variable, 1=dest_x is a constant
// flags,2: 0=dest_y is a variable, 1=dest_y is a constant
// flags,1: 0=size_x is a variable, 1=size_x is a constant
// flags,0: 0=size_y is a variable, 1=size_y is a constant
// if the two sections overlap, the result is undefined
//-------------------------------------------------------------------------------------------
void am_dev_CP_BLK_TR_C( st_am_state * p_state )
{
  unsigned char (* src)[26], (* dest)[26];
  unsigned char flags, transp, src_x, src_y, dest_x, dest_y, size_x, size_y, x, y;
  //get parameters
  flags = p_state->sector[p_state->offset++];
  if( (flags & 0x80) == 0 )
    src = p_state->pixels;
  else
    src = p_state->overlay;
  if( (flags & 0x40) == 0 )
    dest = p_state->pixels;
  else
    dest = p_state->overlay;
  transp = p_state->sector[p_state->offset++];
  src_x = p_state->sector[p_state->offset++];
  if( (flags & 0x20) == 0 )
    src_x = p_state->vars[src_x];
  src_y = p_state->sector[p_state->offset++];
  if( (flags & 0x10) == 0 )
    src_y = p_state->vars[src_y];
  dest_x = p_state->sector[p_state->offset++];
  if( (flags & 0x08) == 0 )
    dest_x = p_state->vars[dest_x];
  dest_y = p_state->sector[p_state->offset++];
  if( (flags & 0x04) == 0 )
    dest_y = p_state->vars[dest_y];
  size_x = p_state->sector[p_state->offset++];
  if( (flags & 0x02) == 0 )
    size_x = p_state->vars[size_x];
  size_y = p_state->sector[p_state->offset++];
  if( (flags & 0x01) == 0 )
    size_y = p_state->vars[size_y];
  //check parameters
  if( src_x >= 26 || src_y >= 20 || dest_x >= 26 || dest_y >= 20 || size_x > 26 || size_y > 20 )
    return;
  if( src_x + size_x > 26 || src_y + size_y > 20 || dest_x + size_x > 26 || dest_y + size_y > 20 )
    return;
  //copy with transparent color
  for( y = 0; y < size_y; y++ )
    for( x = 0; x < size_x; x++ )
      if( src[src_y + y][src_x + x] != transp )
        dest[dest_y + y][dest_x + x] = src[src_y + y][src_x + x];
}

//-------------------------------------------------------------------------------------------
// DISPLAY
// display frame
//-------------------------------------------------------------------------------------------
void am_dev_DISPLAY( st_am_state * p_state )
{
  unsigned char x, y, tmp;
  //we do not have the principle of two drawing buffers in this simulator
  //but we simulate this by placing the old picture back in the drawing buffer
  for( y = 0; y < 20; y++ )
  {
    for( x = 0; x < 26; x++ )
    {
      tmp = p_state->pixels[y][x];
      p_state->pixels[y][x] = p_state->picture[y][x];
      p_state->picture[y][x] = tmp;
    }
  }
}

//-------------------------------------------------------------------------------------------
// DISPLAY_CP
// display frame and copy it into new drawing buffer
//-------------------------------------------------------------------------------------------
void am_dev_DISPLAY_CP( st_am_state * p_state )
{
  unsigned char x, y;
  //we do not have the principle of two drawing buffers in this simulator
  //so we just copy the darwing buffer to the display buffer
  for( y = 0; y < 20; y++ )
    for( x = 0; x < 26; x++ )
      p_state->picture[y][x] = p_state->pixels[y][x];
}

//-------------------------------------------------------------------------------------------
// DELAY_DO <delay_h> <delay_l>
// wait <delay> milliseconds
//-------------------------------------------------------------------------------------------
void am_dev_DELAY_DO( st_am_state * p_state )
{
  unsigned short delay;
  //get delay
  delay = (unsigned short)(p_state->vars[p_state->sector[p_state->offset++]]) << 8;
  delay |= (unsigned short)(p_state->vars[p_state->sector[p_state->offset++]]);
  //get current time
  gettimeofday( &p_state->delay_end, NULL );
  //add delay (milliseconds) to get time when delay ends
  p_state->delay_end.tv_sec += (signed long)(delay / 1000);
  p_state->delay_end.tv_usec += (signed long)(delay % 1000) * 1000;
  if( p_state->delay_end.tv_usec > 1000000 )
    p_state->delay_end.tv_usec -= 1000000, p_state->delay_end.tv_sec++;
  //set flag to wait for delay
  p_state->wait_for_delay = 1;
}

//-------------------------------------------------------------------------------------------
// DELAY_DO_C <cons_delay (2)>
// wait <delay> milliseconds
//-------------------------------------------------------------------------------------------
void am_dev_DELAY_DO_C( st_am_state * p_state )
{
  unsigned short cons_delay;
  //get delay
  cons_delay = (unsigned short)(p_state->sector[p_state->offset++]) << 8;
  cons_delay |= (unsigned short)(p_state->sector[p_state->offset++]);
  //get current time
  gettimeofday( &p_state->delay_end, NULL );
  //add delay (milliseconds) to get time when delay ends
  p_state->delay_end.tv_sec += (signed long)(cons_delay / 1000);
  p_state->delay_end.tv_usec += (signed long)(cons_delay % 1000) * 1000;
  if( p_state->delay_end.tv_usec > 1000000 )
    p_state->delay_end.tv_usec -= 1000000, p_state->delay_end.tv_sec++;
  //set flag to wait for delay
  p_state->wait_for_delay = 1;
}

//-------------------------------------------------------------------------------------------
// DELAY_START <delay_h> <delay_l>
// start background delay of <delay> milliseconds
// - background delay can be checked with DELAY_CHECK
// - background delay can be waited for with DELAY_WAIT
//-------------------------------------------------------------------------------------------
void am_dev_DELAY_START( st_am_state * p_state )
{
  unsigned short delay;
  //get delay
  delay = (unsigned short)(p_state->vars[p_state->sector[p_state->offset++]]) << 8;
  delay |= (unsigned short)(p_state->vars[p_state->sector[p_state->offset++]]);
  //get current time
  gettimeofday( &p_state->delay_end, NULL );
  //add delay (milliseconds) to get time when delay ends
  p_state->delay_end.tv_sec += (signed long)(delay / 1000);
  p_state->delay_end.tv_usec += (signed long)(delay % 1000) * 1000;
  if( p_state->delay_end.tv_usec > 1000000 )
    p_state->delay_end.tv_usec -= 1000000, p_state->delay_end.tv_sec++;
}

//-------------------------------------------------------------------------------------------
// DELAY_START_C <cons_delay (2)>
// start background delay of <delay> milliseconds
// - background delay can be checked with DELAY_CHECK
// - background delay can be waited for with DELAY_WAIT
//-------------------------------------------------------------------------------------------
void am_dev_DELAY_START_C( st_am_state * p_state )
{
  unsigned short cons_delay;
  //get delay
  cons_delay = (unsigned short)(p_state->sector[p_state->offset++]) << 8;
  cons_delay |= (unsigned short)(p_state->sector[p_state->offset++]);
  //get current time
  gettimeofday( &p_state->delay_end, NULL );
  //add delay (milliseconds) to get time when delay ends
  p_state->delay_end.tv_sec += (signed long)(cons_delay / 1000);
  p_state->delay_end.tv_usec += (signed long)(cons_delay % 1000) * 1000;
  if( p_state->delay_end.tv_usec > 1000000 )
    p_state->delay_end.tv_usec -= 1000000, p_state->delay_end.tv_sec++;
}

//-------------------------------------------------------------------------------------------
// DELAY_CHECK <check result>
// check if background delay elapsed (returns 0xFF if delay is active, 0x00 if not)
// - background delay can be started with DELAY_START
// - background delay can be waited for with DELAY_WAIT
//-------------------------------------------------------------------------------------------
void am_dev_DELAY_CHECK( st_am_state * p_state )
{
  struct timeval time;
  unsigned char delay_active;
  //get current time
  gettimeofday( &time, NULL );
  //compare with delay-end-time (watch out: seconds can roll over)
  if( (signed long)(time.tv_sec - p_state->delay_end.tv_sec) < 0 || (time.tv_sec == p_state->delay_end.tv_sec && time.tv_usec < p_state->delay_end.tv_usec) )
    delay_active = 0xFF; //delay still active
  else
    delay_active = 0x00; //delay elapsed
  //write result into output variable
  p_state->vars[p_state->sector[p_state->offset++]] = delay_active;
}

//-------------------------------------------------------------------------------------------
// DELAY_WAIT
// wait for background delay to elapse
// - background delay can be started with DELAY_START
// - background delay can be checked with DELAY_CHECK
//-------------------------------------------------------------------------------------------
void am_dev_DELAY_WAIT( st_am_state * p_state )
{
  //just set flag to wait for delay
  p_state->wait_for_delay = 1;
}

//-------------------------------------------------------------------------------------------
// GET_KEYS <current> <pressed> <released>
// get keys
// - current: current (debounced) key state
// - pressed: keys that were pressed since last GET_KEYS
// - released: keys that were released since last GET_KEYS
//-------------------------------------------------------------------------------------------
void am_dev_GET_KEYS( st_am_state * p_state )
{
  //copy key state into destination variables
  p_state->vars[p_state->sector[p_state->offset++]] = p_state->keys_current;
  p_state->vars[p_state->sector[p_state->offset++]] = p_state->keys_pressed;
  p_state->vars[p_state->sector[p_state->offset++]] = p_state->keys_released;
  //clear pressed and released bits
  p_state->keys_pressed = 0x00;
  p_state->keys_released = 0x00;
}

//-------------------------------------------------------------------------------------------
// GET_SNES0 <current_h> <current_l> <pressed_h> <pressed_l> <released_h> <released_l>
// get keys of 1st SNES pad
// - current: current (debounced) key state
// - pressed: keys that were pressed since last GET_SNES0
// - released: keys that were released since last GET_SNES0
//-------------------------------------------------------------------------------------------
void am_dev_GET_SNES0( st_am_state * p_state )
{
  //copy key state into destination variables
  p_state->vars[p_state->sector[p_state->offset++]] = (unsigned char)((p_state->snes0_current >> 8) & 0xFF);
  p_state->vars[p_state->sector[p_state->offset++]] = (unsigned char)(p_state->snes0_current & 0xFF);
  p_state->vars[p_state->sector[p_state->offset++]] = (unsigned char)((p_state->snes0_pressed >> 8) & 0xFF);
  p_state->vars[p_state->sector[p_state->offset++]] = (unsigned char)(p_state->snes0_pressed & 0xFF);
  p_state->vars[p_state->sector[p_state->offset++]] = (unsigned char)((p_state->snes0_released >> 8) & 0xFF);
  p_state->vars[p_state->sector[p_state->offset++]] = (unsigned char)(p_state->snes0_released & 0xFF);
  //clear pressed and released bits
  p_state->snes0_pressed = 0x0000;
  p_state->snes0_released = 0x0000;
}

//-------------------------------------------------------------------------------------------
// GET_SNES1 <current_h> <current_l> <pressed_h> <pressed_l> <released_h> <released_l>
// get keys of 2nd SNES pad
// - current: current (debounced) key state
// - pressed: keys that were pressed since last GET_SNES1
// - released: keys that were released since last GET_SNES1
//-------------------------------------------------------------------------------------------
void am_dev_GET_SNES1( st_am_state * p_state )
{
  //copy key state into destination variables
  p_state->vars[p_state->sector[p_state->offset++]] = (unsigned char)((p_state->snes1_current >> 8) & 0xFF);
  p_state->vars[p_state->sector[p_state->offset++]] = (unsigned char)(p_state->snes1_current & 0xFF);
  p_state->vars[p_state->sector[p_state->offset++]] = (unsigned char)((p_state->snes1_pressed >> 8) & 0xFF);
  p_state->vars[p_state->sector[p_state->offset++]] = (unsigned char)(p_state->snes1_pressed & 0xFF);
  p_state->vars[p_state->sector[p_state->offset++]] = (unsigned char)((p_state->snes1_released >> 8) & 0xFF);
  p_state->vars[p_state->sector[p_state->offset++]] = (unsigned char)(p_state->snes1_released & 0xFF);
  //clear pressed and released bits
  p_state->snes1_pressed = 0x0000;
  p_state->snes1_released = 0x0000;
}

//-------------------------------------------------------------------------------------------
// GET_TIME <sec> <min> <hour>
// get time: seconds, minutes, hours
//-------------------------------------------------------------------------------------------
void am_dev_GET_TIME( st_am_state * p_state )
{
  time_t time_t_time;
  struct tm tm_time;
  unsigned long secs;
  //get current brokwn-down time and obtain number of seconds since midnight from it
  time( &time_t_time );
  gmtime_r( &time_t_time, &tm_time );
  secs = (unsigned long)tm_time.tm_hour * 3600 + (unsigned long)tm_time.tm_min * 60 + (unsigned long)tm_time.tm_sec;
  //add time offset
  secs += p_state->time_offset;
  //write time into variables
  p_state->vars[p_state->sector[p_state->offset++]] = (unsigned char)(secs % 60);
  p_state->vars[p_state->sector[p_state->offset++]] = (unsigned char)((secs / 60) % 60);
  p_state->vars[p_state->sector[p_state->offset++]] = (unsigned char)((secs / 3600) % 24);
}

//-------------------------------------------------------------------------------------------
// SET_TIME <sec> <min> <hour>
// set time: seconds (0..59), minutes (0..59), hours (0..23)
//-------------------------------------------------------------------------------------------
void am_dev_SET_TIME( st_am_state * p_state )
{
  unsigned long new_secs, secs;
  time_t time_t_time;
  struct tm tm_time;
  //get time from variables
  new_secs = (unsigned long)p_state->vars[p_state->sector[p_state->offset++]];
  new_secs += (unsigned long)p_state->vars[p_state->sector[p_state->offset++]] * 60;
  new_secs += (unsigned long)p_state->vars[p_state->sector[p_state->offset++]] * 3600;
  //get current brokwn-down time and obtain number of seconds since midnight from it
  time( &time_t_time );
  gmtime_r( &time_t_time, &tm_time );
  secs = (unsigned long)tm_time.tm_hour * 3600 + (unsigned long)tm_time.tm_min * 60 + (unsigned long)tm_time.tm_sec;
  //calculate offset and save it
  p_state->time_offset = new_secs + 86400 - secs % 86400; //+ 86400 (= 24h) to avoid negtive values
}

//-------------------------------------------------------------------------------------------
// NOP
// no operation
//-------------------------------------------------------------------------------------------
void am_dev_NOP( st_am_state * p_state )
{
}

void (* am_cmds[0x100])( st_am_state * ) =
{
  /* 0x00 */ am_dev_JMP_NEXT_SEC,
  /* 0x01 */ am_dev_JMP_SEC,
  /* 0x02 */ am_dev_JMP_OFS,
  /* 0x03 */ am_dev_JMP_OFS,
  /* 0x04 */ am_dev_JMP_OFS_SEC,
  /* 0x05 */ am_dev_JMP_OFS_SEC,
  /* 0x06 */ am_dev_JMP_IND,
  /* 0x07 */ am_dev_NOP,
  /* 0x08 */ am_dev_JMP_EQ,
  /* 0x09 */ am_dev_JMP_EQ,
  /* 0x0A */ am_dev_JMP_EQ_C,
  /* 0x0B */ am_dev_JMP_EQ_C,
  /* 0x0C */ am_dev_JMP_NEQ,
  /* 0x0D */ am_dev_JMP_NEQ,
  /* 0x0E */ am_dev_JMP_NEQ_C,
  /* 0x0F */ am_dev_JMP_NEQ_C,
  /* 0x10 */ am_dev_JMP_GT,
  /* 0x11 */ am_dev_JMP_GT,
  /* 0x12 */ am_dev_JMP_GT_C,
  /* 0x13 */ am_dev_JMP_GT_C,
  /* 0x14 */ am_dev_JMP_NGT,
  /* 0x15 */ am_dev_JMP_NGT,
  /* 0x16 */ am_dev_JMP_NGT_C,
  /* 0x17 */ am_dev_JMP_NGT_C,
  /* 0x18 */ am_dev_JMP_LT,
  /* 0x19 */ am_dev_JMP_LT,
  /* 0x1A */ am_dev_JMP_LT_C,
  /* 0x1B */ am_dev_JMP_LT_C,
  /* 0x1C */ am_dev_JMP_NLT,
  /* 0x1D */ am_dev_JMP_NLT,
  /* 0x1E */ am_dev_JMP_NLT_C,
  /* 0x1F */ am_dev_JMP_NLT_C,
  /* 0x20 */ am_dev_JMP_GTS,
  /* 0x21 */ am_dev_JMP_GTS,
  /* 0x22 */ am_dev_JMP_GTS_C,
  /* 0x23 */ am_dev_JMP_GTS_C,
  /* 0x24 */ am_dev_JMP_NGTS,
  /* 0x25 */ am_dev_JMP_NGTS,
  /* 0x26 */ am_dev_JMP_NGTS_C,
  /* 0x27 */ am_dev_JMP_NGTS_C,
  /* 0x28 */ am_dev_JMP_LTS,
  /* 0x29 */ am_dev_JMP_LTS,
  /* 0x2A */ am_dev_JMP_LTS_C,
  /* 0x2B */ am_dev_JMP_LTS_C,
  /* 0x2C */ am_dev_JMP_NLTS,
  /* 0x2D */ am_dev_JMP_NLTS,
  /* 0x2E */ am_dev_JMP_NLTS_C,
  /* 0x2F */ am_dev_JMP_NLTS_C,
  /* 0x30 */ am_dev_JMP_AND_Z,
  /* 0x31 */ am_dev_JMP_AND_Z,
  /* 0x32 */ am_dev_JMP_AND_Z_C,
  /* 0x33 */ am_dev_JMP_AND_Z_C,
  /* 0x34 */ am_dev_JMP_AND_NZ,
  /* 0x35 */ am_dev_JMP_AND_NZ,
  /* 0x36 */ am_dev_JMP_AND_NZ_C,
  /* 0x37 */ am_dev_JMP_AND_NZ_C,
  /* 0x38 */ am_dev_JMP_OR_M,
  /* 0x39 */ am_dev_JMP_OR_M,
  /* 0x3A */ am_dev_JMP_OR_M_C,
  /* 0x3B */ am_dev_JMP_OR_M_C,
  /* 0x3C */ am_dev_JMP_OR_NM,
  /* 0x3D */ am_dev_JMP_OR_NM,
  /* 0x3E */ am_dev_JMP_OR_NM_C,
  /* 0x3F */ am_dev_JMP_OR_NM_C,
  /* 0x40 */ am_dev_NOP,
  /* 0x41 */ am_dev_NOP,
  /* 0x42 */ am_dev_NOP,
  /* 0x43 */ am_dev_NOP,
  /* 0x44 */ am_dev_NOP,
  /* 0x45 */ am_dev_NOP,
  /* 0x46 */ am_dev_NOP,
  /* 0x47 */ am_dev_NOP,
  /* 0x48 */ am_dev_NOP,
  /* 0x49 */ am_dev_NOP,
  /* 0x4A */ am_dev_NOP,
  /* 0x4B */ am_dev_NOP,
  /* 0x4C */ am_dev_NOP,
  /* 0x4D */ am_dev_NOP,
  /* 0x4E */ am_dev_NOP,
  /* 0x4F */ am_dev_NOP,
  /* 0x50 */ am_dev_MOV,
  /* 0x51 */ am_dev_MOV_C,
  /* 0x52 */ am_dev_MOV2,
  /* 0x53 */ am_dev_MOV2_C,
  /* 0x54 */ am_dev_MOV4,
  /* 0x55 */ am_dev_MOV4_C,
  /* 0x56 */ am_dev_NOP,
  /* 0x57 */ am_dev_NOP,
  /* 0x58 */ am_dev_RD_I,
  /* 0x59 */ am_dev_WR_I,
  /* 0x5A */ am_dev_NOP,
  /* 0x5B */ am_dev_NOP,
  /* 0x5C */ am_dev_RD_S,
  /* 0x5D */ am_dev_NOP,
  /* 0x5E */ am_dev_NOP,
  /* 0x5F */ am_dev_NOP,
  /* 0x60 */ am_dev_ADD,
  /* 0x61 */ am_dev_ADD_C,
  /* 0x62 */ am_dev_ADDC,
  /* 0x63 */ am_dev_ADDC_C,
  /* 0x64 */ am_dev_SUB,
  /* 0x65 */ am_dev_SUB_C,
  /* 0x66 */ am_dev_SUBB,
  /* 0x67 */ am_dev_SUBB_C,
  /* 0x68 */ am_dev_MUL,
  /* 0x69 */ am_dev_MUL_C,
  /* 0x6A */ am_dev_MULS,
  /* 0x6B */ am_dev_MULS_C,
  /* 0x6C */ am_dev_NOP,
  /* 0x6D */ am_dev_NOP,
  /* 0x6E */ am_dev_NOP,
  /* 0x6F */ am_dev_NOP,
  /* 0x70 */ am_dev_AND,
  /* 0x71 */ am_dev_AND_C,
  /* 0x72 */ am_dev_OR,
  /* 0x73 */ am_dev_OR_C,
  /* 0x74 */ am_dev_XOR,
  /* 0x75 */ am_dev_XOR_C,
  /* 0x76 */ am_dev_NOP,
  /* 0x77 */ am_dev_NOP,
  /* 0x78 */ am_dev_SHL,
  /* 0x79 */ am_dev_SHL_C,
  /* 0x7A */ am_dev_SHR,
  /* 0x7B */ am_dev_SHR_C,
  /* 0x7C */ am_dev_ROL,
  /* 0x7D */ am_dev_ROL_C,
  /* 0x7E */ am_dev_ROR,
  /* 0x7F */ am_dev_ROR_C,
  /* 0x80 */ am_dev_NOP,
  /* 0x81 */ am_dev_NOP,
  /* 0x82 */ am_dev_NOP,
  /* 0x83 */ am_dev_NOP,
  /* 0x84 */ am_dev_NOP,
  /* 0x85 */ am_dev_NOP,
  /* 0x86 */ am_dev_NOP,
  /* 0x87 */ am_dev_NOP,
  /* 0x88 */ am_dev_NOP,
  /* 0x89 */ am_dev_NOP,
  /* 0x8A */ am_dev_NOP,
  /* 0x8B */ am_dev_NOP,
  /* 0x8C */ am_dev_NOP,
  /* 0x8D */ am_dev_NOP,
  /* 0x8E */ am_dev_NOP,
  /* 0x8F */ am_dev_NOP,
  /* 0x90 */ am_dev_RD_PIX,
  /* 0x91 */ am_dev_RD_PIX_C,
  /* 0x92 */ am_dev_RD_OVP,
  /* 0x93 */ am_dev_RD_OVP_C,
  /* 0x94 */ am_dev_WR_PIX,
  /* 0x95 */ am_dev_WR_PIX_C,
  /* 0x96 */ am_dev_WR_OVP,
  /* 0x97 */ am_dev_WR_OVP_C,
  /* 0x98 */ am_dev_WR_PIX_CO,
  /* 0x99 */ am_dev_WR_PIX_CO_C,
  /* 0x9A */ am_dev_WR_OVP_CO,
  /* 0x9B */ am_dev_WR_OVP_CO_C,
  /* 0x9C */ am_dev_NOP,
  /* 0x9D */ am_dev_NOP,
  /* 0x9E */ am_dev_NOP,
  /* 0x9F */ am_dev_NOP,
  /* 0xA0 */ am_dev_WR_PIX_LINE,
  /* 0xA1 */ am_dev_WR_PIX_LINE_C,
  /* 0xA2 */ am_dev_WR_OVP_LINE,
  /* 0xA3 */ am_dev_WR_OVP_LINE_C,
  /* 0xA4 */ am_dev_WR_PIX_QUART,
  /* 0xA5 */ am_dev_WR_PIX_QUART_C,
  /* 0xA6 */ am_dev_WR_OVP_QUART,
  /* 0xA7 */ am_dev_WR_OVP_QUART_C,
  /* 0xA8 */ am_dev_WR_PIX_CLR,
  /* 0xA9 */ am_dev_WR_OVP_CLR,
  /* 0xAA */ am_dev_WR_PIX_SET,
  /* 0xAB */ am_dev_WR_OVP_SET,
  /* 0xAC */ am_dev_WR_PIX_BW,
  /* 0xAD */ am_dev_WR_OVP_BW,
  /* 0xAE */ am_dev_WR_PIX_SEL,
  /* 0xAF */ am_dev_WR_PIX_SEL_C,
  /* 0xB0 */ am_dev_WR_OVP_SEL,
  /* 0xB1 */ am_dev_WR_OVP_SEL_C,
  /* 0xB2 */ am_dev_NOP,
  /* 0xB3 */ am_dev_NOP,
  /* 0xB4 */ am_dev_CP_BLK,
  /* 0xB5 */ am_dev_NOP,
  /* 0xB6 */ am_dev_CP_BLK_TR,
  /* 0xB7 */ am_dev_CP_BLK_TR_C,
  /* 0xB8 */ am_dev_NOP,
  /* 0xB9 */ am_dev_NOP,
  /* 0xBA */ am_dev_NOP,
  /* 0xBB */ am_dev_NOP,
  /* 0xBC */ am_dev_NOP,
  /* 0xBD */ am_dev_NOP,
  /* 0xBE */ am_dev_NOP,
  /* 0xBF */ am_dev_NOP,
  /* 0xC0 */ am_dev_DISPLAY,
  /* 0xC1 */ am_dev_DISPLAY_CP,
  /* 0xC2 */ am_dev_NOP,
  /* 0xC3 */ am_dev_NOP,
  /* 0xC4 */ am_dev_NOP,
  /* 0xC5 */ am_dev_NOP,
  /* 0xC6 */ am_dev_NOP,
  /* 0xC7 */ am_dev_NOP,
  /* 0xC8 */ am_dev_DELAY_DO,
  /* 0xC9 */ am_dev_DELAY_DO_C,
  /* 0xCA */ am_dev_DELAY_START,
  /* 0xCB */ am_dev_DELAY_START_C,
  /* 0xCC */ am_dev_DELAY_CHECK,
  /* 0xCD */ am_dev_DELAY_WAIT,
  /* 0xCE */ am_dev_NOP,
  /* 0xCF */ am_dev_NOP,
  /* 0xD0 */ am_dev_GET_KEYS,
  /* 0xD1 */ am_dev_NOP,
  /* 0xD2 */ am_dev_GET_SNES0,
  /* 0xD3 */ am_dev_GET_SNES1,
  /* 0xD4 */ am_dev_GET_TIME,
  /* 0xD5 */ am_dev_SET_TIME,
  /* 0xD6 */ am_dev_NOP,
  /* 0xD7 */ am_dev_NOP,
  /* 0xD8 */ am_dev_NOP,
  /* 0xD9 */ am_dev_NOP,
  /* 0xDA */ am_dev_NOP,
  /* 0xDB */ am_dev_NOP,
  /* 0xDC */ am_dev_NOP,
  /* 0xDD */ am_dev_NOP,
  /* 0xDE */ am_dev_NOP,
  /* 0xDF */ am_dev_NOP,
  /* 0xE0 */ am_dev_NOP,
  /* 0xE1 */ am_dev_NOP,
  /* 0xE2 */ am_dev_NOP,
  /* 0xE3 */ am_dev_NOP,
  /* 0xE4 */ am_dev_NOP,
  /* 0xE5 */ am_dev_NOP,
  /* 0xE6 */ am_dev_NOP,
  /* 0xE7 */ am_dev_NOP,
  /* 0xE8 */ am_dev_NOP,
  /* 0xE9 */ am_dev_NOP,
  /* 0xEA */ am_dev_NOP,
  /* 0xEB */ am_dev_NOP,
  /* 0xEC */ am_dev_NOP,
  /* 0xED */ am_dev_NOP,
  /* 0xEE */ am_dev_NOP,
  /* 0xEF */ am_dev_NOP,
  /* 0xF0 */ am_dev_NOP,
  /* 0xF1 */ am_dev_NOP,
  /* 0xF2 */ am_dev_NOP,
  /* 0xF3 */ am_dev_NOP,
  /* 0xF4 */ am_dev_NOP,
  /* 0xF5 */ am_dev_NOP,
  /* 0xF6 */ am_dev_NOP,
  /* 0xF7 */ am_dev_NOP,
  /* 0xF8 */ am_dev_NOP,
  /* 0xF9 */ am_dev_NOP,
  /* 0xFA */ am_dev_NOP,
  /* 0xFB */ am_dev_NOP,
  /* 0xFC */ am_dev_NOP,
  /* 0xFD */ am_dev_NOP,
  /* 0xFE */ am_dev_NOP,
  /* 0xFF */ am_dev_NOP
};

//-------------------------------------------------------------------------------------------
// initialize an ARCADEmini state
// returns 0 on success
//-------------------------------------------------------------------------------------------
int am_dev_init_state( st_am_state * p_state, FILE * p_file )
{
  long long pos;
  char sig_buf[32], signature[32] = "ARCADEmini 26x20/16 ambin 1.0   ";

  //check signature
  pos = 0;
  if( fseeko64( p_file, pos, SEEK_SET ) != 0 )
    return -1;
  if( fread( sig_buf, 1, sizeof( sig_buf ), p_file ) != 32 )
    return -1;
  if( memcmp( sig_buf, signature, sizeof( sig_buf ) ) != 0 )
    return -1;

  //assign file
  p_state->p_file = p_file;
  //start at begin of second sector
  p_state->sector_no = 1;
  p_state->offset = 0;
  //load sector
  am_dev_load_sector( p_state );
  //clear overflow buffer
  memset( p_state->sector_ov, 0, sizeof( p_state->sector_ov ) );
  //clear registers
  memset( p_state->vars, 0, sizeof( p_state->vars ) );
  //clear pixel buffer, overlay-buffer, picture
  memset( p_state->pixels, 0, sizeof( p_state->pixels ) );
  memset( p_state->overlay, 0, sizeof( p_state->overlay ) );
  memset( p_state->picture, 0, sizeof( p_state->picture ) );
  //no delay active
  gettimeofday( &p_state->delay_end, NULL );
  p_state->wait_for_delay = 0;
  //clear state of keys
  p_state->keys_current = 0x00;
  p_state->keys_pressed = 0x00;
  p_state->keys_released = 0x00;
  //clear state of snes keys
  p_state->snes0_current = 0x0000;
  p_state->snes0_pressed = 0x0000;
  p_state->snes0_released = 0x0000;
  p_state->snes1_current = 0x0000;
  p_state->snes1_pressed = 0x0000;
  p_state->snes1_released = 0x0000;
  //no time offset - use system time
  p_state->time_offset = 0;
  //success
  return 0;
}

//-------------------------------------------------------------------------------------------
// execute an ARCADEmini command
// (returns number of milliseconds left of current delay or 0 if not waiting for a delay)
//-------------------------------------------------------------------------------------------
unsigned short am_dev_exec( st_am_state * p_state )
{
  struct timeval time;
  signed long sec, usec;
  unsigned char cmd;

  //do not execute next command before end of delay
  if( p_state->wait_for_delay )
  {
    //get current time
    gettimeofday( &time, NULL );
    //calculate time to wait
    sec = p_state->delay_end.tv_sec - time.tv_sec;
    usec = p_state->delay_end.tv_usec - time.tv_usec;
    if( usec < 0 )
      usec += 1000000, sec--;
    //check seconds and convert into sinlge microseconds value
    if( sec > 100 ) //this should not occur, because the delay was set using a unsigned short milliseconds value
      sec = 100; //and 65536 milliseconds are much less than 100 seconds
    if( sec >= 0 ) //delay still active
      usec += sec * 1000000;
    else //delay is over, so just use any negative value
      usec = -1;
    //delay still active
    if( usec > 0 )
      //return number of millisoconds left of delay
      return (unsigned short)((usec > 65535000)?(65535):(usec / 1000));
    //delay ist over, so clear flag
    p_state->wait_for_delay = 0;
  }

  //read command
  cmd = p_state->sector[p_state->offset++];
  //execute command
  am_cmds[cmd]( p_state );  
  //return no delay to wait for
  return 0;
}

//-------------------------------------------------------------------------------------------
// set a new state of the 8 keys
//-------------------------------------------------------------------------------------------
void am_dev_keystate( st_am_state * p_state, unsigned char keystate )
{
  //update pressed keys (were not down, are now down)
  p_state->keys_pressed |= (~p_state->keys_current & keystate);
  //update released keys (were down, are now not down)
  p_state->keys_released |= (p_state->keys_current & ~keystate);
  //update current keystate
  p_state->keys_current = keystate;
}

//-------------------------------------------------------------------------------------------
// set a new state of the snes keys
//-------------------------------------------------------------------------------------------
void am_dev_snes_keystate( st_am_state * p_state, unsigned char snes_no, unsigned short keystate )
{
  //SNES0
  if( snes_no == 0 )
  {
    //update pressed keys (were not down, are now down)
    p_state->snes0_pressed |= (~p_state->snes0_current & keystate);
    //update released keys (were down, are now not down)
    p_state->snes0_released |= (p_state->snes0_current & ~keystate);
    //update current keystate
    p_state->snes0_current = keystate;
  }
  //SNES1
  if( snes_no == 1 )
  {
    //update pressed keys (were not down, are now down)
    p_state->snes1_pressed |= (~p_state->snes1_current & keystate);
    //update released keys (were down, are now not down)
    p_state->snes1_released |= (p_state->snes1_current & ~keystate);
    //update current keystate
    p_state->snes1_current = keystate;
  }
}

